import pdb

import numpy as np

def precision(ranked_label_list, k):
    return np.mean(ranked_label_list[:,:k].sum(1) / k)

def recall(ranked_label_list, k):
    return (ranked_label_list[:,:k].sum(1) / ranked_label_list.sum(1)).mean()

def ndcg(ranked_label_list, k=None):
    if k is None: k = ranked_label_list.shape[1]
    discount = np.log2(np.arange(2,2+k))
    dcg = np.sum(ranked_label_list[:,:k] / discount, 1)
    idcg = np.sum(np.flip(np.sort(ranked_label_list.copy()), 1)[:,:k] / discount,1)
    ndcg = (dcg / idcg).mean()
    return ndcg



'''
Implement Logistic Regression model for tabular individual outcome
prediction in clinical trials.
'''
import pdb
import os
import joblib
import pickle

from sklearn.linear_model import LogisticRegression as lr_model

from pytrial.data.patient_data import TabularPatientBase
from pytrial.utils.check import check_checkpoint_file, check_model_dir, check_model_config_file, make_dir_if_not_exist
from .base import TabularIndivBase

class BuildModel:
    def __new__(self, config):
        if config['dual']:
            solver = 'liblinear'
        else:
            solver = 'lbfgs'
        model = lr_model(
            C = 1/config['weight_decay'],
            dual = config['dual'],
            solver = solver,
            max_iter = config['epochs'],
            )
        return model

class LogisticRegression(TabularIndivBase):
    '''
    Implement Logistic Regression model for tabular individual outcome
    prediction in clinical trials. Now only support `binary classification`.

    Parameters
    ----------
    weigth_decay: float
        Regularization strength for l2 norm; must be a positive float.
        Like in support vector machines, smaller values specify weaker regularization.

    dual: bool
        Dual or primal formulation. Dual formulation is only implemented for l2 penalty with liblinear solver.
        Prefer `dual=False` when `n_samples > n_features`.

    epochs: int
        Maximum number of iterations taken for the solvers to converge.

    experiment_id: str, optional (default='test')
        The name of current experiment. Decide the saved model checkpoint name.
    '''
    def __init__(self,
        weight_decay=1,
        dual=False,
        epochs=100,
        experiment_id='test',
        ) -> None:
        super().__init__(experiment_id=experiment_id)
        self.config = {
            'weight_decay': weight_decay,
            'dual': dual,
            'epochs': epochs,
            'experiment_id': experiment_id,
            'model_name': 'logistic_regression',
        }
        self._save_config(self.config)

    def fit(self, train_data, valid_data=None):
        '''Train logistic regression model to predict patient outcome
        with tabular input data.

        Parameters
        ----------
        train_data: dict
            {

            'x': TabularPatientBase or pd.DataFrame,

            'y': pd.Series or np.ndarray

            }

            - 'x' contain all patient features; 
            - 'y' contain labels for each row.

        valid_data: Ignored.
            Not used, present heare for API consistency by convention.
        '''
        self._input_data_check(train_data)
        self._build_model()

        if isinstance(train_data['x'], TabularPatientBase):
            dataset = train_data['x']
            x_feat = dataset.df
            y = train_data['y']
        else:
            x_feat = train_data['x']
            y = train_data['y']

        self._fit_model(x_feat, y)

    def predict(self, test_data):
        '''
        Make prediction probability based on the learned model.
        Save to `self.result_dir`.

        Parameters
        ----------
        test_data: dict
            {

            'x': TabularPatientBase or pd.DataFrame,

            'y': pd.Series or np.ndarray

            }

            - 'x' contain all patient features; 
            - 'y' contain labels for each row. Ignored for prediction function.

        Returns
        -------
        ypred: np.ndarray
            The predicted probability for each patient.
            
            - For binary classification, return shape (n, );
            - For multiclass classification, return shape (n, n_class).

        '''
        self._input_data_check(test_data)

        dataset = test_data['x']
        if isinstance(dataset, TabularPatientBase):
            x_feat = dataset.df
        else:
            x_feat = dataset

        ypred = self.model.predict_proba(x_feat)
        if ypred.shape[1] == 2: # binary
            ypred = ypred[:,1]

        # save results to dir
        pickle.dump(ypred, open(os.path.join(self.result_dir, 'pred.pkl'), 'wb'))
        if isinstance(test_data, dict):
            if 'y' in test_data:
                pickle.dump(test_data['y'], open(os.path.join(self.result_dir, 'label.pkl'), 'wb'))

        return ypred

    def save_model(self, output_dir=None):
        '''
        Save the learned logistic regression model to the disk.

        Parameters
        ----------
        output_dir: str or None
            The dir to save the learned model.
            If set None, will save model to `self.checkout_dir`.
        '''
        if output_dir is not None:
            make_dir_if_not_exist(output_dir)
        else:
            output_dir = self.checkout_dir

        self._save_config(self.config, output_dir=output_dir)
        ckpt_path = os.path.join(output_dir, 'indiv-tabular.model')
        joblib.dump(self.model, ckpt_path)

    def load_model(self, checkpoint=None):
        '''
        Save the learned logistic regression model to the disk.

        Parameters
        ----------
        checkpoint: str or None
            - If a directory, the only checkpoint file `.model` will be loaded.
            - If a filepath, will load from this file;
            - If None, will load from `self.checkout_dir`.
        '''
        if checkpoint is None:
            checkpoint = self.checkout_dir

        checkpoint_filename = check_checkpoint_file(checkpoint, suffix='model')
        config_filename = check_model_config_file(checkpoint)
        self.model = joblib.load(checkpoint_filename)
        self.config = self._load_config(config_filename)

    def _build_model(self):
        self.model = BuildModel(self.config)

    def _fit_model(self, x_feat, y):
        self.model.fit(x_feat, y)



import pdb
from copy import deepcopy

import promptehr # third-part promptEHR package

from .base import SequenceSimulationBase
from ..data import SequencePatient

class BuildModel:
    def __new__(self, config):
        model = promptehr.PromptEHR(
            **config,
            )
        return model

class PromptEHR(SequenceSimulationBase):
    '''Implement a BART equiped with prompt learning for synthetic patient records simulation [1]_.

    Parameters
    ----------
    code_type: list[str]
        A list of code types that the model will learn and generate. For example, ``code_type=['diag','prod','med']``.

    token_dict: dict[list]
        A dictionary of new tokens (code events, e.g., ICD code) that the model needs to learn and generate.
        For example, ``token_dict={'diag':['D01','D02','D03'], 'prod':['P01','P02','P03'], 'med':['M01','M02','M03']}``.

    n_num_feature: int
        Number of numerical patient baseline features. Notice that it assumes that the input
        baseline features are **ALWAYS** numerical feature first. That is to say,
        the input baseline ``feature = [num1, num2, .., num_n, cat1, cat2,...]``.

    cat_cardinalities: list[int]
        The number of categories for each categorical patient baseline features.
        The input baseline ``feature = [num1, num2, .., num_n, cat1, cat2,...]``.

    epochs: int
        Num training epochs in total.

    batch_size: int
        Training batch size.

    eval_batch_size: int
        Evaluation batch size.
    
    evaluation_steps: int
        How many steps of updates then try to evaluate the trained models.
    
    learning_rate: float
        Training learning rate.
    
    weight_decay: float
        Training weight decay.
    
    num_worker: int
        Numer of dataloading paralleled processes.
    
    output_dir: str
        Training logs output to this folder.

    device: str or list[int]
        Should be str like `cuda:0` or `cpu`, otherwise should be a list GPU ids.
    
    Notes
    -----
    .. [1] Wang, Z. and Sun, J. (2022). PromptEHR: Conditional Electronic Healthcare Records Generation with Prompt Learning. EMNLP 2022.
    '''
    model = None
    sample_config = {
        'num_beams': 1, # >1: beam_sample; =1: sample_gen
        'no_repeat_ngram_size': 1,
        'do_sample': True,
        'num_return_sequences': 1,
        'code_type': '',
        'top_k': 1,
        'temperature': 1.0,
        'max_length': 6,
    }
    def __init__(self,
        code_type=None,
        n_num_feature=None,
        cat_cardinalities=None,
        epochs=50,
        batch_size=16,
        eval_batch_size=16,
        evaluation_steps=1000,
        learning_rate=5e-5,
        weight_decay=1e-4,
        num_worker=8,
        device='cuda:0',
        experiment_id='trial_simulation.sequence.promptehr',
        ):
        super().__init__(experiment_id)
        self.config = {
            'code_type': code_type,
            'n_num_feature':n_num_feature,
            'cat_cardinalities':cat_cardinalities,
            'epoch':epochs,
            'batch_size':batch_size,
            'eval_batch_size':eval_batch_size,
            'eval_step':evaluation_steps,
            'learning_rate':learning_rate,
            'weight_decay':weight_decay,
            'num_worker':num_worker,
            'output_dir':self.checkout_dir,
            'device':device,
        }
        config = deepcopy(self.config)
        self.model = BuildModel(config)
        self.device = device

    def fit(self, train_data, val_data=None):
        '''
        Fit PromptEHR model on the input training EHR data.

        Parameters
        ----------
        train_data: SequencePatient
            A `SequencePatient` contains patient records where 'v' corresponds to 
            visit sequence of different events.

        val_data: SequencePatient
            A `SequencePatient` contains patient records where 'v' corresponds to 
            visit sequence of different events.
        '''
        self.model.fit(train_data=train_data, val_data=val_data)

    def predict(self, test_data, n_per_sample=None, n=None, sample_config=None, verbose=True):
        '''
        Generate synthetic records based on input real patient seq data.

        Parameters
        ----------
        test_data: SequencePatient
            A `SequencePatient` contains patient records where 'v' corresponds to 
            visit sequence of different events.
        
        n: int
            How many samples in total will be generated.
        
        n_per_sample: int
            How many samples generated based on each indivudals.

        sample_config: dict
            Configuration for sampling synthetic records, key parameters:

            - 'num_beams': Number of beams in beam search, if set `1` then beam search is deactivated;
            - 'top_k': Sampling from top k candidates.
            - 'temperature': temperature to make sampling distribution flater or skewer.
        
        verbose: bool
            If print the progress bar or not.

        Returns
        -------
        fake_data: SequencePatient
            Synthetic patient records in `SequencePatient` format.
        '''
        return_res = self.model.predict(
            test_data=test_data, 
            n_per_sample=n_per_sample, 
            n=n, 
            sample_config=sample_config,
            verbose=verbose)

        fake_metadata = deepcopy(test_data.metadata)
        fake_data = {
            'v':return_res['visit'], 
            'x':return_res['feature'],
            }
        if 'y' in return_res:
            fake_data['y'] = return_res['y']
        fake_data = SequencePatient(data=fake_data, metadata=fake_metadata)
        return fake_data

    def load_model(self, checkpoint):
        '''
        Load model and the pre-encoded trial embeddings from the given
        checkpoint dir.

        Parameters
        ----------
        checkpoint: str
            The input dir that stores the pretrained model.
        '''
        self.model.load_model(checkpoint)

    def save_model(self, output_dir):
        '''
        Save the learned simulation model to the disk.

        Parameters
        ----------
        output_dir: str
            The dir to save the learned model.
        '''
        self.model.save_model(output_dir)

    def evaluate(self, test_data):
        '''
        Evaluate the perplexity of trained PromptEHR model on the input data, will test the perplexity
        for each type of codes.
        
        Parameters
        ----------
        test_data: PatientSequence
            Standard sequential patient records in `PatientSequence` format.
        '''
        self.model.evaluate(test_data)

    def from_pretrained(self, input_dir='./simulation/pretrained_promptEHR'):
        '''
        Load pretrained PromptEHR model and make patient EHRs generation.
        Pretrained model was learned from MIMIC-III patient sequence data.

        Parameters
        ----------
        input_dir: str
            The path to the pretrained model. If no found, will download from online and save
            to this folder.
        '''
        self.model.from_pretrained(input_dir=input_dir)
        self.config.update(self.model.config)
    
    def update_config(self, config):
        '''
        Update the model configuration. Will be useful when load pretrained model
        and want to finetune on new data.

        Parameters
        ----------
        config: dict
            The configuration for the model.
        '''
        self.config.update(config)
        self.model.update_config(config)


name = 'PyTrial'
version = '0.0.5'

import random
import numpy as np
import torch

def manual_seed(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False


# save all constant variables used by the package
CODE_TYPES = ['tbd']
SPECIAL_TOKEN_DICT = {'tbd':['<tbd>','</tbd>']}

UNKNOWN_TOKEN = '<unk>'

model_max_length = 512

eps = 1e-16

# PRETRAINED_MODEL_URL = 'https://uofi.box.com/shared/static/cu09as2bmotrr9bejsfgumv6yx46mmfw.zip'
PRETRAINED_MODEL_URL = 'https://storage.googleapis.com/pytrial/promptEHR_pretrained.zip'

SYNTHETIC_DATA_URL = 'https://github.com/RyanWangZf/PromptEHR/raw/main/demo_data/synthetic_ehr/data.pkl'

# a name mapping from the original promptehr config to the training_args
config_to_train_args = {
    'epochs': 'num_train_epochs',
    'num_worker': 'dataloader_num_workers',
    'batch_size': 'per_device_train_batch_size',
    'eval_batch_size': 'per_device_eval_batch_size',
    'eval_step': 'eval_steps',
}


'''
Provide data functions for trial data simulation.
'''
from collections import defaultdict
from pytrial.data.trial_data import TrialDatasetStructured, TrialDataCollator
from pytrial.data.site_data import SiteBase,  SiteBaseDemographics, ModalitySiteBase
from torch.utils.data import Dataset
import numpy as np
import torch

class TrialSiteSimple(Dataset):
    '''
    Load trial-site dataset with trial and site features.
    Each corresponds to a set of historical enrollment
    values for training trial site selection models.

    Parameters
    ----------
    site_data: df
        A df used to generate site data
    trial_data: pd.DataFrame
        A dataframe containing trial information including a set of paired trials named `label` and `enrollment` 
        which contains a mapping to sites in the site data that participated in the trial and the corresponding
        enrollments of those sites.
    '''
    def __init__(self, site_data, trial_data):
        self.sites = SiteBaseDemographics(site_data) if 'demographics' in site_data else SiteBase(site_data)
        self.mappings = trial_data.pop('label')
        self.enrollments = trial_data.pop('enrollment')
        self.trials = TrialDatasetStructured(trial_data)
        
    def __len__(self):
        return len(self.trials)

    def __getitem__(self, idx):
        sitesPerTrial = self.mappings[idx]
        sample = {"trial": self.trials[idx], 
                  "site": [self.sites[int(s)] for s in sitesPerTrial],
                  "label": self.enrollments[idx], 
                  "eth_label": None if isinstance(self.sites, SiteBase) else [self.sites.get_label(sites) for sites in sitesPerTrial],
                 }
        return sample
    
class TrialSiteModalities(Dataset):
    '''
    Load trial-site dataset with trial and site features.
    Each corresponds to a set of historical enrollment
    values for training trial site selection models.

    Parameters
    ----------
    site_data: df
        A df used to generate site data
    trial_data: pd.DataFrame
        A dataframe containing trial information including a set of paired trials named `label` and `enrollment` 
        which contains a mapping to sites in the site data that participated in the trial and the corresponding
        enrollments of those sites.
    '''
    def __init__(self, site_data, trial_data):
        self.sites = ModalitySiteBase(site_data)
        self.mappings = trial_data.pop('label')
        self.enrollments = trial_data.pop('enrollment')
        self.trials = TrialDatasetStructured(trial_data)

    def __len__(self):
        return len(self.trials)

    def __getitem__(self, idx):
        sitesPerTrial = self.mappings[idx]
        siteFeatures = [self.sites[trialSites] for trialSites in sitesPerTrial]
        siteFeatures = {k: np.stack([d[k] for d in siteFeatures]) for k in siteFeatures[0]}
        sample = {"trial": self.trials[idx], 
                  "label": self.enrollments[idx], 
                  "eth_label": siteFeatures['eth_label'],
                  "inv_static": siteFeatures['x'],
                  "dx": siteFeatures['dx'],
                  "dx_len": siteFeatures['dx_lens'],
                  "rx": siteFeatures['rx'],
                  "rx_len": siteFeatures['rx_lens'],
                  "enroll_hist": siteFeatures['hist'],
                  "enroll_hist_len": siteFeatures['hist_lens'],
                  "inv_mask": np.zeros((len(sitesPerTrial), 4), dtype=np.int8)
                 }
        return sample

class SiteSelectionBaseCollator:
    '''
    support make collation of unequal sized list of batch for densely stored
    sequential visits data.

    Parameters
    ----------
    config: dict
        
        {

        'has_demographics': in 'true' or 'false'

        }
        
    '''
    has_demographics = False
    config = {'has_demographics':'False'} 

    def __init__(self, config=None):
        if config is not None:
            self.config.update(config)
            
        if str(self.config['has_demographics']) == 'True': self.has_demographics = True
        else: self.has_demographics = False

    def __call__(self, inputs):
        '''
        Paramters
        ---------
        inputs = {
            'trial': list[np.ndarray], 
        
            'site': list[list[np.ndarray]],
            
            'label': list[list[int]],
            
            'eth_label' None or list[list[np.ndarray]]
        }

        Returns
        -------
        outputs = {
            'trial': tensor, 
        
            'site': tensor,
            
            'label': tensor,
            
            'eth_label' None or tensor
        }
        '''
        # init output dict
        return_data = defaultdict(list)

        return_data['trial'] = torch.FloatTensor(np.array([i['trial'] for i in inputs])).squeeze(1)
        return_data['site'] = torch.FloatTensor(np.array([i['site'] for i in inputs])).squeeze(2)
        return_data['label'] = torch.FloatTensor(np.array([i['label'] for i in inputs]))

        # processing all
        if self.has_demographics:
            return_data['eth_label'] = torch.FloatTensor(np.array([i['eth_label'] for i in inputs]))

        return return_data
    
class SiteSelectionModalitiesCollator:
    '''
    support make collation of unequal sized list of batch for densely stored
    sequential visits data.

    Parameters
    ----------
    config: dict
        
        {

        'visit_mode': in 'dense' or 'tensor',

        'trial_mode': in 'dense' or 'tensor',
        
        'has_demographics': in 'true' or 'false',

        }
        
    '''
    is_tensor_visit = False
    is_tensor_trial = False
    has_demographics = False
    config = {'visit_mode':'dense', 'trial_mode':'dense', 'has_demographics':'false'} 

    def __init__(self, config=None):
        if config is not None:
            self.config.update(config)
            
        if self.config['visit_mode'] == 'tensor': self.is_tensor_visit = True
        else: self.is_tensor_visit = False

        if self.config['trial_mode'] == 'tensor': self.is_tensor_trial = True
        else: self.is_tensor_trial = False
        
        if self.config['has_demographics'] == 'true': self.has_demographics = True
        else: self.has_demographics = False

    def __call__(self, inputs):
        '''
        Paramters
        ---------
        inputs = {
            'trial': list[np.ndarray], 
            
            'label': list[list[int]],
            
            'eth_label' None or list[list[np.ndarray]]
            
            'inv_static': list[list[np.ndarray]]
            
            'dx': list[list[np.ndarray]]
            
            'dx_len': list[list[int]]
            
            'rx': list[list[np.ndarray]]
            
            'rx_len': list[list[int]]
            
            'enroll_hist': list[list[np.ndarray]]
            
            'enroll_hist_len': list[list[int]]
            
            'inv_mask': list[list[np.ndarray]]
        }

        Returns
        -------
        outputs = {
            'trial': tensor,
            
            'label': tensor,
            
            'eth_label' None or tensor,
            
            'inv_static': tensor
            
            'dx': tensor
            
            'dx_len': tensor
            
            'rx': tensor
            
            'rx_len': tensor
            
            'enroll_hist': tensor
            
            'enroll_hist_len': tensor
            
            'inv_mask': tensor
        }
        '''
        
        # init output dict
        return_data = defaultdict(list)
        
        inputs = {k: np.stack([d[k] for d in inputs]) for k in inputs[0]}
        
        return_data['trial'] = torch.FloatTensor(np.array(inputs['trial']))
        return_data['label'] = torch.FloatTensor(np.array(inputs['label']))
        return_data['inv_static'] = torch.FloatTensor(np.array(inputs['inv_static']))
        return_data['dx'] = torch.FloatTensor(np.array(inputs['dx']))
        return_data['dx_len'] = torch.LongTensor(np.array(inputs['dx_len']))
        return_data['rx'] = torch.FloatTensor(np.array(inputs['rx']))
        return_data['rx_len'] = torch.LongTensor(np.array(inputs['rx_len']))
        return_data['enroll_hist'] = torch.FloatTensor(np.array(inputs['enroll_hist']))
        return_data['enroll_hist_len'] = torch.LongTensor(np.array(inputs['enroll_hist_len']))
        return_data['inv_mask'] = torch.LongTensor(np.array(inputs['inv_mask']))

        # processing all
        if self.has_demographics:
            return_data['eth_label'] = torch.FloatTensor(np.array(inputs['eth_label']))
        
        return return_data


import os
import pdb
import joblib
import warnings

import numpy as np
import pandas as pd
import torch
from torch import nn
from torch.nn import BatchNorm1d, Linear, Module, Sequential
from torch.nn.functional import cross_entropy, mse_loss, sigmoid
from torch.optim import Adam
from torch.utils.data import DataLoader, TensorDataset

from pytrial.data.patient_data import TabularPatientBase
# from pytrial.utils.check import check_checkpoint_file, check_model_dir, check_model_config_file, make_dir_if_not_exist
from pytrial.utils.tabular_utils import get_transformer
from pytrial.utils.check import check_checkpoint_file, check_model_dir, check_model_config_file, make_dir_if_not_exist


from .base import TabularSimulationBase

class ResidualFC(Module):
    def __init__(self, input_dim, output_dim, activate, bn_decay):
        super(ResidualFC, self).__init__()
        self.seq = Sequential(
            Linear(input_dim, output_dim),
            BatchNorm1d(output_dim, momentum=bn_decay),
            activate()
        )

    def forward(self, input):
        residual = self.seq(input)
        return input + residual


class Generator(Module):
    def __init__(self, random_dim, hidden_dim, bn_decay):
        super(Generator, self).__init__()

        dim = random_dim
        seq = []
        for item in list(hidden_dim)[:-1]:
            assert item == dim
            seq += [ResidualFC(dim, dim, nn.ReLU, bn_decay)]
        assert hidden_dim[-1] == dim
        seq += [
            Linear(dim, dim),
            BatchNorm1d(dim, momentum=bn_decay),
            nn.ReLU()
        ]
        self.seq = Sequential(*seq)

    def forward(self, input):
        return self.seq(input)


class Discriminator(Module):
    def __init__(self, data_dim, hidden_dim):
        super(Discriminator, self).__init__()
        dim = data_dim * 2
        seq = []
        for item in list(hidden_dim):
            seq += [
                Linear(dim, item),
                nn.ReLU() if item > 1 else nn.Sigmoid()
            ]
            dim = item
        self.seq = Sequential(*seq)

    def forward(self, input):
        mean = input.mean(dim=0, keepdim=True)
        mean = mean.expand_as(input)
        inp = torch.cat((input, mean), dim=1)
        return self.seq(inp)


class Encoder(Module):
    def __init__(self, data_dim, compress_dims, embedding_dim):
        super(Encoder, self).__init__()
        dim = data_dim
        seq = []
        for item in list(compress_dims) + [embedding_dim]:
            seq += [
                Linear(dim, item),
                nn.ReLU()
            ]
            dim = item
        self.seq = Sequential(*seq)

    def forward(self, input):
        return self.seq(input)


class Decoder(Module):
    def __init__(self, embedding_dim, decompress_dims, data_dim):
        super(Decoder, self).__init__()
        dim = embedding_dim
        seq = []
        for item in list(decompress_dims):
            seq += [
                Linear(dim, item),
                nn.ReLU()
            ]
            dim = item
        seq.append(Linear(dim, data_dim))
        self.seq = Sequential(*seq)

    def forward(self, input, output_info):
        return self.seq(input)


def aeloss(fake, real, output_info):
    st = 0
    loss = []
    for item in output_info:
        if item[1] == 'sigmoid':
            ed = st + item[0]
            loss.append(mse_loss(sigmoid(fake[:, st:ed]), real[:, st:ed], reduction='sum'))
            st = ed
        elif item[1] == 'softmax':
            ed = st + item[0]
            loss.append(cross_entropy(
                fake[:, st:ed], torch.argmax(real[:, st:ed], dim=-1), reduction='sum'))
            st = ed
        else:
            assert 0
    return sum(loss) / fake.size()[0]


class MedGANSynthesizer:
    """docstring for MedGAN."""

    def __init__(
        self,
        embedding_dim=128,
        random_dim=128,
        generator_dims=(128, 128),          # 128 -> 128 -> 128
        discriminator_dims=(256, 128, 1),   # datadim * 2 -> 256 -> 128 -> 1
        compress_dims=(),                   # datadim -> embedding_dim
        decompress_dims=(),                 # embedding_dim -> datadim
        bn_decay=0.99,
        l2scale=0.001,
        pretrain_epoch=200,
        batch_size=1000,
        epochs=2000,
        device='cpu',
        verbose=False,
        ):

        self.embedding_dim = embedding_dim
        self.random_dim = random_dim
        self.generator_dims = generator_dims
        self.discriminator_dims = discriminator_dims

        self.compress_dims = compress_dims
        self.decompress_dims = decompress_dims
        self.bn_decay = bn_decay
        self.l2scale = l2scale

        self.pretrain_epoch = pretrain_epoch
        self.batch_size = batch_size
        self.epochs = epochs

        self.device = device
        self.verbose = verbose

    def _get_metadata(self, data):
        self.output_info = []
        for k, v in data.metadata['transformed_col2col'].items():
            if len(v) == 1:
                self.output_info.append(
                    (1, 'sigmoid')
                )
            else:
                self.output_info.append(
                    (len(v), 'softmax')
                )

    def fit(self, data):
        data_val = data.df.values
        dataset = TensorDataset(torch.from_numpy(data_val.astype('float32')).to(self.device))
        loader = DataLoader(dataset, batch_size=self.batch_size, shuffle=True, drop_last=False)
        
        self._get_metadata(data)
        
        # data_dim = self.transformer.output_dim
        data_dim = data_val.shape[1]
        encoder = Encoder(data_dim, self.compress_dims, self.embedding_dim).to(self.device)
        self.decoder = Decoder(self.embedding_dim, self.compress_dims, data_dim).to(self.device)
        optimizerAE = Adam(
            list(encoder.parameters()) + list(self.decoder.parameters()),
            weight_decay=self.l2scale
        )

        for i in range(self.pretrain_epoch):
            if self.verbose:
                print('Pretrain Epoch: {} / {}'.format(i+1, self.pretrain_epoch))
            for id_, data in enumerate(loader):
                optimizerAE.zero_grad()
                real = data[0].to(self.device)
                emb = encoder(real)
                rec = self.decoder(emb, self.output_info)
                loss = aeloss(rec, real, self.output_info)

                loss.backward()
                optimizerAE.step()

        self.generator = Generator(
            self.random_dim, self.generator_dims, self.bn_decay).to(self.device)
        discriminator = Discriminator(data_dim, self.discriminator_dims).to(self.device)
        optimizerG = Adam(
            list(self.generator.parameters()) + list(self.decoder.parameters()),
            weight_decay=self.l2scale
        )
        optimizerD = Adam(discriminator.parameters(), weight_decay=self.l2scale)

        mean = torch.zeros(self.batch_size, self.random_dim, device=self.device)
        std = mean + 1
        bs_larger_than_n_sample = False
        for i in range(self.epochs):
            n_d = 2
            n_g = 1
            
            if len(loader) < n_d:
                # batch size is larger than all samples
                # so we have to update discriminator each iteration
                bs_larger_than_n_sample = True
                warnings.warn(
                    f"""
                    Batch size is larger than all samples.
                    Discriminator will be updated each iteration instead of each {n_d} iterations.
                    Consider using a smaller batch size.
                    """
                )
            
            for id_, data in enumerate(loader):
                real = data[0].to(self.device)
                noise = torch.normal(mean=mean, std=std)
                emb = self.generator(noise)
                fake = self.decoder(emb, self.output_info)

                optimizerD.zero_grad()
                y_real = discriminator(real)
                y_fake = discriminator(fake)
                real_loss = -(torch.log(y_real + 1e-4).mean())
                fake_loss = (torch.log(1.0 - y_fake + 1e-4).mean())
                loss_d = real_loss - fake_loss
                loss_d.backward()
                optimizerD.step()

                if id_ % n_d == 0 or bs_larger_than_n_sample:
                    for _ in range(n_g):
                        noise = torch.normal(mean=mean, std=std)
                        emb = self.generator(noise)
                        fake = self.decoder(emb, self.output_info)
                        optimizerG.zero_grad()
                        y_fake = discriminator(fake)
                        loss_g = -(torch.log(y_fake + 1e-4).mean())
                        loss_g.backward()
                        optimizerG.step()

            if self.verbose:
                print(f'epoch {i} loss_d: {loss_d.item()} loss_g: {loss_g.item()}')


    def sample(self, n):
        self.generator.eval()
        self.decoder.eval()

        steps = n // self.batch_size + 1
        data = []
        for i in range(steps):
            mean = torch.zeros(self.batch_size, self.random_dim)
            std = mean + 1
            noise = torch.normal(mean=mean, std=std).to(self.device)
            emb = self.generator(noise)
            fake = self.decoder(emb, self.output_info)
            fake = torch.sigmoid(fake)
            data.append(fake.detach().cpu().numpy())
        data = np.concatenate(data, axis=0)
        data = data[:n]
        return data
    
# utils
# ------------

# main functions
# ------------

class BuildModel:
    def __new__(self, config) -> MedGANSynthesizer:
        model = MedGANSynthesizer(
            embedding_dim=config['embedding_dim'],
            random_dim=config['random_dim'],
            generator_dims=config['generator_dims'],
            discriminator_dims=config['discriminator_dims'],
            compress_dims=config['compress_dims'],
            decompress_dims=config['decompress_dims'],
            bn_decay=config['bn_decay'],
            l2scale=config['l2scale'],
            pretrain_epoch=config['pretrain_epoch'],
            epochs=config['epochs'],
            batch_size=config['batch_size'],
            device=config['device'],
            verbose=config['verbose'],
        )
        return model


class MedGAN(TabularSimulationBase):
    '''
    Implement MedGAN model for patient level tabular data generation [1]_.

    Parameters
    ----------
    embedding_dim : int, default 128
        Dimension of embedding layer.
    
    random_dim : int, default 128
        Dimension of random noise.
    
    generator_dims : tuple, default (128, 128)
        Dimension of generator layers.
    
    discriminator_dims : tuple, default (256, 128, 1)
        Dimension of discriminator layers.
    
    compress_dims : tuple, default ()
        Dimension of compressed embedding layer. datadim -> embedding_dim
    
    decompress_dims : tuple, default ()
        Dimension of decompressed embedding layer. embedding_dim -> datadim
    
    bn_decay : float, default 0.99
        Decay rate of batch normalization.
    
    l2scale : float, default 0.001
        L2 regularization scale.
    
    pretrain_epoch : int, default 200
        Number of pretrain epochs.
    
    batch_size : int, default 1000
        Batch size for training.
    
    epochs : int, default 1000
        Number of epochs for training.

    experiment_id: str
        Experiment id for logging.

    verbose: bool
        Whether to print training information.
    
    Notes
    -----
    .. [1] Choi, E., Biswal, S., Malin, B., Duke, J., Stewart, W. F., & Sun, J. (2017, November). Generating multi-label discrete patient records using generative adversarial networks. In Machine learning for healthcare conference (pp. 286-305). PMLR.
    '''
    def __init__(self,
        embedding_dim=128,
        random_dim=128,
        generator_dims=(128, 128),          # 128 -> 128 -> 128
        discriminator_dims=(256, 128, 1),   # datadim * 2 -> 256 -> 128 -> 1
        compress_dims=(),                   # datadim -> embedding_dim
        decompress_dims=(),                 # embedding_dim -> datadim
        bn_decay=0.99,
        l2scale=0.001,
        pretrain_epoch=200,
        batch_size=1000,
        epochs=2000,
        device='cpu',
        experiment_id='trial_simulation.tabular.medgan',
        verbose=False,
        ):
        super().__init__(experiment_id=experiment_id)
        self.config = {
            'embedding_dim': embedding_dim,
            'random_dim': random_dim,
            'generator_dims': generator_dims,
            'discriminator_dims': discriminator_dims,
            'compress_dims': compress_dims,
            'decompress_dims': decompress_dims,
            'bn_decay': bn_decay,
            'l2scale': l2scale,
            'pretrain_epoch': pretrain_epoch,
            'batch_size': batch_size,
            'epochs': epochs,
            'device': device,
            'verbose':verbose,
        }
        self._save_config(self.config)
    
    def fit(self, train_data):
        '''
        Train MedGAN model to generate synthetic tabular patient data.

        Parameters
        ----------
        train_data : TabularPatientBase
            Training data.
        '''
        self._input_data_check(train_data)
        self._build_model()
        self._fit_model(train_data)
    
    def predict(self, n):
        '''
        Generate synthetic tabular patient data.

        Parameters
        ----------
        n : int
            Number of samples to generate.
        
        Returns
        -------
        data : np.ndarray
            Generated synthetic data.
        '''
        return self.model.sample(n)
    
    def load_model(self, checkpoint):
        '''
        Load model from checkpoint.

        Parameters
        ----------
        checkpoint : str
            Path to checkpoint.
            If a directory is given, will load the latest checkpoint in the directory.
            If a filepath is given, will load the checkpoint from the filepath.
            If set None, will load from default directory `self.checkpoint_dir`.
        '''
        if checkpoint is None:
            checkpoint = self.checkpoint_dir
        
        checkpoint_filename = check_checkpoint_file(checkpoint, suffix='model')
        config_filename = check_model_config_file(checkpoint)
        self.model = joblib.load(checkpoint_filename)
        self.config = self._load_config(config_filename)

    def save_model(self, output_dir):
        '''
        Save model to checkpoint.

        Parameters
        ----------
        output_dir : str
            Output directory. If set None, will save to default directory `self.checkpoint_dir`.
        '''
        if output_dir is None:
            output_dir = self.checkpoint_dir

        make_dir_if_not_exist(output_dir)
        self._save_config(self.config, output_dir=output_dir)
        ckpt_path = os.path.join(output_dir, 'ctgan.model')
        joblib.dump(self.model, ckpt_path)

    def _build_model(self):
        self.model = BuildModel(self.config)

    def _fit_model(self, dataset):
        self.model.fit(dataset)



import pdb

import torch
from torch import nn
import torch.nn.functional as F
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import DataLoader
import numpy as np


from pytrial.utils.check import (
    check_checkpoint_file, check_model_dir, check_model_config_file, make_dir_if_not_exist
)
from .base import PatientTrialMatchBase
from ..data import TrialCollator, PatientCollator
from ..losses import COMPOSELoss
from ..trainer import PatientTrialTrainer


class COMPOSE(PatientTrialMatchBase):
    '''
    Leverage COMPOSE model for patient-trial matching [1]_.

    One patient's label = [[0,1,2,3], [2,3,4,5]] where
    the first is a list of indices of inclusion criteria that the patient satisfies.
    the second is a list of indices of exclusion criteria that the patient does not satisfy.
    for the other criteria we dont know if the patient satisfied or not.
    In this regard, we need to:
    (1) predict "match" for all inclusion criteria of a trial for recruiting the patient
    (2) predict "unmatch" for all exclusion criteria of a trial for recruiting the patient

    # prediction logits
    # 0 is unmatch
    # 1 is match
    # 2 is unknown

    Parameters
    ----------
    order: list[str]
        The order of events within each visit in the patient's EHR data, e.g., orders=['diag','med','prod'].

    vocab_size: int
        The vocabular size of each patient's EHR event types, e.g., diag, med, prod.

    max_visit: int
        Maximum number of visits to load for EHRs inputs.

    word_dim: int
        The dimension of input word embeddings for encoding eligibility criteria.

    conv_dim: int
        The dimension of convolutional layers for processing eligibility criteria embeddings.

    mem_dim: int
        The hidden dimension of the EHR memory network (encode patient EHRs).

    mlp_dim: int
        The hidden dimension of the MLP layers in the Query Network.

    demo_dim: int
        The input dimensions of patient demographic information, e.g., age, gender.

    margin: float
        The margin when compute nn.CosineEmbeddingLoss on patient embedding and inclusion/exclusion criteria embeddings.
        Refer to Eq. (12) of the reference paper.
    
    epochs: int, optional (default=50)
        Number of iterations (epochs) over the training data.

    batch_size: int, optional (default=512)
        Number of samples in each training batch.

    learning_rate: float, optional (default=1e-3)
        The learning rate.
    
    weight_decay: float (default=0)
        The weight decay during training.

    num_worker: int
        Number of workers used to do dataloading during training.

    device: str
        The model device.
    
    Notes
    -----
    .. [1] Gao, J., et al. (2020, August). COMPOSE: cross-modal pseudo-siamese network for patient trial matching. In Proceedings of the 26th ACM SIGKDD International Conference on Knowledge Discovery & Data Mining (pp. 803-812).
    '''
    def __init__(self,
        order,
        vocab_size,
        max_visit=20,
        word_dim=768,
        conv_dim=128,
        mem_dim=320,
        mlp_dim=512,
        demo_dim=3,
        margin=0.3,
        batch_size=512,
        epochs=50,
        learning_rate=1e-3,
        weight_decay=0,
        num_worker=0,
        device='cuda:0',
        experiment_id='test',
        ) -> None:
        super().__init__(experiment_id)
        self.config = {
            'order':order,
            'max_visit':max_visit,
            'vocab_size':vocab_size,
            'word_dim':word_dim,
            'conv_dim':conv_dim,
            'mem_dim':mem_dim,
            'mlp_dim':mlp_dim,
            'demo_dim':demo_dim,
            'margin':margin,
            'batch_size':batch_size,
            'learning_rate':learning_rate,
            'weight_decay':weight_decay,
            'epochs':epochs,
            'num_worker':num_worker,
            'device':device,
            }
        self.device = device
        self.config['total_vocab_size'] = sum(vocab_size)
        self._build_model()
    
    def predict(self, test_data, return_dict=False):
        '''
        Predict the matching of patient and ECs of each target trial.
        Will make predictions for each patient, go through all included trials.
        For example, given 1000 patients and 10 trials, will generate predictions
        as the shape of ``1000 x 10 x num_eligibility_criteria``.

        Parameters
        ----------
        test_data: dict
            A dict contains patient and trial data, respecitvely.
            
            {

            'patient': `pytrial.tasks.trial_patient_match.data.PatientData`,

            'trial': `pytrial.tasks.trial_patient_match.data.TrialData`

            }
        
        return_dict: bool
            If return results stored in dictionary, otherwise return tuples of results

        '''
        self._input_data_check(test_data)
        patient_dataloader, trial_dataloader = self.get_test_dataloader(test_data)
        pred_res = self._predict_on_dataloader(patient_dataloader, trial_dataloader)

        if return_dict:
            pred_res['pred_trial'] = self._tuple_result_to_dict(pred_res['pred_trial'])

        return pred_res

    def fit(self, train_data, valid_data=None):
        '''
        Fit patient-trial matching model.

        Parameters
        ----------
        train_data: dict
            A dict contains patient and trial data, respectively.

            {

            'patient': `pytrial.tasks.trial_patient_match.data.PatientData`,

            'trial': `pytrial.tasks.trial_patient_match.data.TrialData`

            }

        valid_data: dict
            A dict contains patient and trial data for evaluation. Same format as
            `train_data`.

        '''
        self._input_data_check(train_data)
        if valid_data is not None: self._input_data_check(valid_data)
        self._fit_model(train_data, valid_data)

    def save_model(self, output_dir):
        '''
        Save the learned patient-match model to the disk.

        Parameters
        ----------
        output_dir: str or None
            The dir to save the learned model.
            If set None, will save model to `self.checkout_dir`.
        '''
        if output_dir is not None:
            make_dir_if_not_exist(output_dir)
        else:
            output_dir = self.checkout_dir
        self._save_config(self.config, output_dir=output_dir)
        self._save_checkpoint({'model':self.model}, output_dir=output_dir)


    def load_model(self, checkpoint):
        '''
        Load model and the pre-encoded trial embeddings from the given
        checkpoint dir.

        Parameters
        ----------
        checkpoint: str
            The input dir that stores the pretrained model.
            If a directory, the only checkpoint file `*.pth.tar` will be loaded.
            If a filepath, will load from this file.
        '''
        checkpoint_filename = check_checkpoint_file(checkpoint)
        config_filename = check_model_config_file(checkpoint)
        state_dict = torch.load(checkpoint_filename)
        if config_filename is not None:
            config = self._load_config(config_filename)
            self.config.update(config)
        self.model = state_dict['model']


    def get_train_dataloader(self, train_data):
        trial_data = train_data['trial']
        patient_data = train_data['patient']

        patient_dataloader = DataLoader(patient_data,
            batch_size=self.config['batch_size'],
            num_workers=self.config['num_worker'],
            pin_memory=True,
            shuffle=True,
            collate_fn=PatientCollator(
                config={
                    'visit_mode':train_data['patient'].metadata['visit']['mode'],
                    'label_mode':train_data['patient'].metadata['label']['mode'],
                    }
                ),
            )
        
        trial_dataloader = DataLoader(trial_data,
            batch_size=self.config['batch_size'],
            num_workers=self.config['num_worker'],
            pin_memory=True,
            shuffle=True,
            collate_fn=TrialCollator(),
        )

        return patient_dataloader, trial_dataloader


    def get_test_dataloader(self, test_data):
        trial_data = test_data['trial']
        patient_data = test_data['patient']

        patient_dataloader = DataLoader(patient_data,
            batch_size=self.config['batch_size'],
            num_workers=self.config['num_worker'],
            pin_memory=False,
            shuffle=False,
            collate_fn=PatientCollator(
                config={
                    'visit_mode':patient_data.metadata['visit']['mode'],
                    'label_mode':patient_data.metadata['label']['mode'],
                    }
                ),
            )
        
        trial_dataloader = DataLoader(trial_data,
            batch_size=1, # get one trial once when making predictions
            num_workers=self.config['num_worker'],
            pin_memory=False,
            shuffle=False,
            collate_fn=TrialCollator(),
        )

        return patient_dataloader, trial_dataloader

    def _build_model(self):
        self.model = BuildModel(
            total_vocab_size=self.config['total_vocab_size'],
            word_dim=self.config['word_dim'],
            conv_dim=self.config['conv_dim'],
            mem_dim=self.config['mem_dim'],
            demo_dim=self.config['demo_dim'],
            mlp_dim=self.config['mlp_dim'],
        )
        if 'cuda' in self.device:
            self.model.cuda()

    def _fit_model(self, train_data, valid_data):
        patient_dataloader, trial_dataloader = self.get_train_dataloader(train_data)
        loss_models = self._build_loss_model()
        # need to customize the training process in trainer
        train_objectives = [[[patient_dataloader,trial_dataloader], loss_model] for loss_model in loss_models]
        trainer = PatientTrialTrainer(model=self,
            train_objectives=train_objectives,
            test_data=valid_data,        
        )
        trainer.train(**self.config)
    
    @torch.no_grad()
    def _predict_on_dataloader(self, patient_dataloader, trial_dataloader, return_label=False):
        '''Return trial level matching predictions.
        Return label for evaluation if required.
        '''
        self.eval()

        # record predictions
        pred_all, label_all = [], []
        pred_inc, pred_exc = [], []

        # record labels
        label_trial_inc, label_trial_exc = [], []
        label_patient_inc, label_patient_exc = [], []

        for i, batch_trial in enumerate(trial_dataloader):
            
            pred_patient_inc, pred_patient_exc = [], []

            label_trial = []
            label_trial_inc_, label_trial_exc_ = [], []

            for batch_ehr in patient_dataloader:
                inputs = {}
                inputs.update(batch_ehr)
                inputs.update(batch_trial)
                inputs = self._prepare_input(inputs)
                pred = self.model(inputs)
                pred_patient_inc.append(pred['logit_inc'])
                pred_patient_exc.append(pred['logit_exc'])

                if return_label:
                    # match trial ec idnex with patient index
                    match_inc, match_inc_ec = self._match_patient_with_trial_ec(batch_trial=batch_trial, batch_ehr=batch_ehr, ec='inc', ec_idx=0)
                    match_exc, match_exc_ec = self._match_patient_with_trial_ec(batch_trial=batch_trial, batch_ehr=batch_ehr, ec='exc', ec_idx=1)
                    match_label = match_inc * ~match_exc
                    label_trial.append(match_label.astype(int))
                    label_trial_inc_.append(match_inc_ec) # bs_patient, num_ec
                    label_trial_exc_.append(match_exc_ec)

                    if i == 0:
                        # get patient label
                        [label_patient_inc.append(y[0]) for y in batch_ehr['y']]
                        [label_patient_exc.append(y[1]) for y in batch_ehr['y']]


            if return_label:
                label_trial_inc_ = np.concatenate(label_trial_inc_)
                label_trial_exc_ = np.concatenate(label_trial_exc_)
                label_trial_inc.append(label_trial_inc_)
                label_trial_exc.append(label_trial_exc_)
                label_all.append(np.concatenate(label_trial))

            # get a batch of trial to all EHRs matching
            pred_patient_inc = torch.cat(pred_patient_inc, 1) # 1, num_patients, num_inc_ec, output_class
            pred_patient_exc = torch.cat(pred_patient_exc, 1) # 1, num_patients, num_exc_ec, output_class

            pred_trial_inc = self._match_trial_for_patients(pred_patient_inc, 'inc')
            pred_trial_exc = self._match_trial_for_patients(pred_patient_exc, 'exc')
            pred_trial = pred_trial_exc * pred_trial_inc # iff both inc and exc are matched, shape=(bs_trial, bs_patient)

            pred_all.append((batch_trial['nct_id'], pred_trial.detach().cpu().numpy()))
            pred_inc.append(pred_patient_inc.detach().cpu().numpy())
            pred_exc.append(pred_patient_exc.detach().cpu().numpy())        
        
        return_res = {
            'pred_trial':pred_all, # trial-level prediction
            'pred_inc': pred_inc, # inclusion-level prediction
            'pred_exc': pred_exc, # exclusion-level prediction
        }

        if return_label:
            return_res['label_trial'] = np.stack(label_all) # trial-level labels
            return_res['label_patient_inc'] = label_patient_inc # patient inclusion criteria label
            return_res['label_patient_exc'] = label_patient_exc # patient exclusion criteria label
            return_res['label_trial_inc'] = label_trial_inc # trial inclusion criteria index
            return_res['label_trial_exc'] = label_trial_exc # trial exclusion criteria index

        return return_res

    def _build_loss_model(self):
        return [COMPOSELoss(self.model)]

    def _prepare_input(self, data):
        '''
        Prepare inputs for compose model encoding.
        {
            'v': patient visit events,
            'x': patient features,
            'y': which ECs patient are matched,
            'inc_ec_index': sampled batch of inclusion criteria index,
            'exc_ec_index': sampled batch of exclusion criteria index,
            'inc_ec_index': sampled batch of inclusion criteria embedding,
            'exc_ec_index': sampled batch of exclusion criteria embedding,
        }
        '''
        inputs = self._prepare_ehr_input(data=data)
        trial_inputs = self._prepare_trial_patient_match_input(data=data)
        inputs.update(trial_inputs)
        return inputs

    def _prepare_ehr_input(self, data):
        inputs = {}
        visits = data['v']
        feature = data['x']

        if not isinstance(feature, torch.Tensor): feature = torch.tensor(feature)
        feature = feature.to(self.device)
        v_lengths = [len(visits[self.config['order'][0]][idx][:self.config['max_visit']]) for idx in range(len(visits[self.config['order'][0]]))]
        mask = torch.zeros(len(v_lengths), max(v_lengths))

        v = torch.zeros(len(v_lengths), max(v_lengths), self.config['total_vocab_size'])
        for idx in range(len(v_lengths)):
            v[idx,:v_lengths[idx]] = torch.tensor(self._translate_dense_visits_to_sparse({k: visits[k][idx][:self.config['max_visit']] for k in visits}))
            mask[idx,:v_lengths[idx]] = 1

        if 'cuda' in self.device:
            v = v.cuda()
            mask = mask.cuda()
            feature = feature.cuda()

        inputs['x'] = feature.float()
        inputs['mask'] = mask
        inputs['v'] = v
        return inputs

    def _prepare_trial_patient_match_input(self, data):
        outputs = {}
        inc_inputs = self._prepare_ec_input(data, 'inc', 0) # patient label 0 corresponds to inclusion
        exc_inputs = self._prepare_ec_input(data, 'exc', 1) # patient label 1 corresponds to exclusion
        outputs.update(inc_inputs)
        outputs.update(exc_inputs)
        return outputs

    def _prepare_ec_input(self, data, ec, label_idx):
        patient_label = data['y']
        inc_label = []
        inc_idx = data[f'{ec}_ec_index']
        inc_emb = data[f'{ec}_ec_emb']
        max_n_inc = [len(l) for l in inc_idx]
        inc_mask = torch.zeros(len(max_n_inc), max(max_n_inc))

        num_patient = len(patient_label)
        num_sample = min(len(max_n_inc), num_patient)

        inc_emb_ts = []
        for idx in range(num_sample): 
            # go through all trials
            # if len(patients) > len(trials)
            # only take the top len(trials) samples
            ec_embs = torch.zeros(max(max_n_inc), inc_emb[0].shape[-1])
            ec_embs[:max_n_inc[idx]] = inc_emb[idx]
            inc_emb_ts.append(ec_embs)
            inc_mask[idx][:max_n_inc[idx]] = 1
            p_label = patient_label[idx][label_idx]
            
            # each trial match each patient
            # trial1 -> patient1
            # trial2 -> patient2
            # ...
            # trial10 -> patient10
            # the remaining of patients will be ignored if len(trial) == 10
            inc_label_ = np.in1d(inc_idx[idx], p_label)

            if ec == 'inc': # inclusion
                inc_label_ = torch.tensor(inc_label_.astype(int)) # match is 1
                inc_label_[inc_label_==0]  = 2 # for the other inclusion it is unknown
            else: # exclusion
                inc_label_ = torch.tensor((~inc_label_).astype(int)) # unmatch is 0
                inc_label_[inc_label_!=0] = 2 # for the other exclusion it is unknown

            if 'cuda' in self.device: inc_label_ = inc_label_.cuda()
            inc_label.append(inc_label_)

        inc_emb_ts = torch.stack(inc_emb_ts)

        if 'cuda' in self.device:
            inc_emb_ts = inc_emb_ts.cuda()
            inc_mask = inc_mask.cuda()

        inc_label = pad_sequence(inc_label, batch_first=True)

        outputs = {
            f'{ec}_emb': inc_emb_ts,
            f'{ec}_emb_mask': inc_mask,
            f'{ec}_label': inc_label,
            }
        return outputs
    
    def _match_patient_with_trial_ec(self, batch_trial, batch_ehr, ec, ec_idx):
        trial_inc_idx = batch_trial[f'{ec}_ec_index']# assume test dataloader only has 1 sample each batch
        patient_label = batch_ehr['y']
        match_label, match_ec_label = [], []
        for p_label in patient_label:
            matched = np.in1d(trial_inc_idx, p_label[ec_idx]).astype(int)
            
            if ec_idx == 0: # inclusion
                label = sum(matched) == len(trial_inc_idx)
                matched[matched==0] = 2 # unk
            else: # exclusion
                label = sum(matched) == 0
                matched[matched==0] = 2 # unk
                matched[matched==1] = 0 # unmatch
            
            match_ec_label.append(matched)
            match_label.append(label)

        return np.array(match_label), np.stack(match_ec_label)

class BuildModel(nn.Module):
    '''
    Build COMPOSE model.
    '''
    def __init__(self, 
        total_vocab_size,
        word_dim, 
        conv_dim,
        mem_dim,
        demo_dim,
        mlp_dim,
        ) -> None:
        super().__init__()
        self.embedding_network = ECEmbedding(word_dim=word_dim, conv_dim=conv_dim)
        self.memory_network = EHRMemoryNetwork(total_vocab_size=total_vocab_size, word_dim=word_dim, mem_dim=mem_dim, demo_dim=demo_dim)
        self.query_network = QueryNetwork(mem_dim=mem_dim, conv_dim=conv_dim, mlp_dim=mlp_dim)

    def forward(self, inputs):
        ehr = inputs['v'] # tensor bs, n_visit, n_event
        demo = inputs['x'] # tensor bs, n_feature
        ehr_mask = inputs['mask']
        memory = self.memory_network(ehr, demo, ehr_mask) # batch_size, 2, mem_dim

        inc_ec, inc_ec_mask = inputs['inc_emb'], inputs['inc_emb_mask']
        inc_emb, inc_ec_emb = self.embedding_network(inc_ec, inc_ec_mask) # bs, num_ec, 512

        exc_ec, exc_ec_mask = inputs['exc_emb'], inputs['exc_emb_mask']
        exc_emb, exc_ec_emb = self.embedding_network(exc_ec, exc_ec_mask) # bs, num_ec, 512

        ouptut_inc, response_inc, query_inc = self.query_network(memory, inc_ec_emb, inc_ec_mask) # bs_trial, bs_ehr, num_ec, 2
        output_exc, response_exc, query_exc = self.query_network(memory, exc_ec_emb, exc_ec_mask) # bs_trial, bs_ehr, num_ec, 2

        # pred_inc = torch.softmax(ouptut_inc, dim=-1)
        # pred_exc = torch.softmax(output_exc, dim=-1)
        return {'logit_inc':ouptut_inc, 
                'logit_exc':output_exc, 
                'response_inc':response_inc,
                'response_exc':response_exc,
                'query_inc':query_inc,
                'query_exc':query_exc,
                }


class CausalConv1d(torch.nn.Conv1d):
    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 stride=1,
                 dilation=1,
                 groups=1,
                 bias=True):

        super(CausalConv1d, self).__init__(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=0,
            dilation=dilation,
            groups=groups,
            bias=bias)
        
        self.__padding = (kernel_size - 1) * dilation
        
    def forward(self, input):
        return super(CausalConv1d, self).forward(F.pad(input, (self.__padding, 0)))

class HighwayBlock(nn.Module):
    def __init__(self, input_dim, kernel_size):
        super(HighwayBlock, self).__init__()
        self.conv_t = CausalConv1d(input_dim, input_dim, kernel_size)
        self.conv_z = CausalConv1d(input_dim, input_dim, kernel_size)

    def forward(self, input):
        t = torch.sigmoid(self.conv_t(input))
        z = t * self.conv_z(input) + (1-t) * input
        return z
    
class ECEmbedding(nn.Module):
    def __init__(self, word_dim, conv_dim):
        super(ECEmbedding, self).__init__()
        # TODO
        # add trainable embeddings: ec_emb, ec_emb_fixed
        #Input: batch_size * sentence_len * embd_dim
        self.word_dim = word_dim
        self.conv_dim = conv_dim
        
        #High-way network for word features
        self.init_conv1 = CausalConv1d(word_dim, conv_dim, 1)
        self.init_conv2 = CausalConv1d(word_dim, conv_dim, 3)
        self.init_conv3 = CausalConv1d(word_dim, conv_dim, 5)
        self.init_conv4 = CausalConv1d(word_dim, conv_dim, 7)
        self.highway1 = HighwayBlock(4*conv_dim, 3)
        self.highway2 = HighwayBlock(4*conv_dim, 3)
        self.highway3 = HighwayBlock(4*conv_dim, 3)

        self.pool = nn.AdaptiveMaxPool1d(1)
    
    def forward(self, input, mask):
        #Input: B * L * embd_dim
        #Mask: B * L

        input = input.permute(0,2,1)
        conv1 = self.init_conv1(input)
        conv2 = self.init_conv2(input)
        conv3 = self.init_conv3(input)
        conv4 = self.init_conv4(input)
        concat = torch.cat((conv1,conv2,conv3,conv4), dim=1)

        highway_res = self.highway1(concat)
        highway_res = torch.relu(highway_res)

        highway_res = self.highway2(highway_res)
        highway_res = torch.relu(highway_res)
        highway_res = self.highway3(highway_res)
        highway_res = torch.relu(highway_res) # bs, 512, num_ec

        highway_res = highway_res * mask.unsqueeze(1) # bs, 512, num_ec
        pooled_res = self.pool(highway_res)
        pooled_res = pooled_res.squeeze(-1) # bs, 512, encoding for the whole trial?
        highway_res = highway_res.permute(0,2,1) # bs, num_ec, 512
        return pooled_res, highway_res
  
class EHRMemoryNetwork(nn.Module):
    def __init__(self, total_vocab_size, word_dim, mem_dim, demo_dim):
        super(EHRMemoryNetwork, self).__init__()
        self.mem_dim = mem_dim
        self.ehr_embedding_matrix = nn.Linear(total_vocab_size, word_dim, bias=False)
        self.erase_layer = nn.Linear(word_dim, mem_dim)
        self.add_layer = nn.Linear(word_dim, mem_dim)
        self.demo_embd = nn.Linear(demo_dim, mem_dim)

        self.init_memory = nn.Parameter(torch.randn(1, mem_dim))

    def forward(self, input, demo, mask):
        '''
        TODO: use hierarchy of codes to enhance code embeddings
        Now: we only make multihot embedding of the input codes.
        input is ehr concept embedding with (bs, num_visit, 12=3x4, word_dim),
        12 is 3 types of codes and 4 layers of code hierarchy,
        maxpooling is applied to input EHRs when each type has multiple codes
        '''
        batch_size = input.size(0)
        time_step = input.size(1)
        memory = self.init_memory
        demo_mem = torch.tanh(self.demo_embd(demo.float()))

        for i in range(time_step):
            cur_input = input[:, i, ...] # bs, vocab_size
            cur_input = self.ehr_embedding_matrix(cur_input) # bs, word_dim
            erase = torch.sigmoid(self.erase_layer(cur_input)) # bs, 320
            add = torch.tanh(self.add_layer(cur_input)) # bs, 320
            cur_mask = mask[:, i].reshape(batch_size, 1) # bs, 1
            erase = erase * cur_mask
            add = add * cur_mask
            update_ratio = (1 - erase) + add # bs, 320
            memory = memory * update_ratio # bs, 320
        
        memory = torch.stack([memory, demo_mem], dim=1) # bs, 2, mem_dim
        return memory

class QueryNetwork(nn.Module):
    def __init__(self, mem_dim, conv_dim, mlp_dim):
        super(QueryNetwork, self).__init__()
        self.word_trans = nn.Linear(4*conv_dim,mem_dim, bias=False)
        self.mlp = nn.Linear(2*mem_dim, mlp_dim)
        self.output = nn.Linear(mlp_dim, 3) # 0: match, 1: umatch, 2: unknown
  
    def forward(self, memory, query, ec_mask):
        #query: bs, num_ec, 4*conv_dim
        #memory: bs, 2, mem_dim

        trans_query = self.word_trans(query) 
        trans_query = torch.relu(trans_query) #bs, num_ec, mem_dim

        # bs_trial, bs, num_ec, mem_dim
        trans_query = trans_query[None].repeat(len(memory), 1, 1, 1).permute(1,0,2,3)
        
        pred_list = []
        response_list = []
        # go through all trials
        for idx, t_query in enumerate(trans_query):
            # mask is only useful when computing loss?
            mask = ec_mask[idx] # num_ec

            # t_query.shape = bs, num_ec, mem_dim
            attention = torch.bmm(t_query, memory.permute(0,2,1)) # bs, num_ec, 2
            attention = torch.softmax(attention, dim=-1)  # bs, num_ec, 2, attention over memory slots
            response = torch.bmm(attention, memory) # bs, num_ec, mem_dim

            output = torch.cat((response, t_query), dim=-1) # bs, num_ec, 2*mem_dim
            output = self.mlp(output) # bs, num_ec, 512
            output = torch.relu(output)
            output = self.output(output) # bs, num_ec, 2 or 3(if label has the class `neutral`)

            pred_list.append(output)
            response_list.append(response)
        
        pred_list = torch.stack(pred_list, 0) # bs_trial, bs_ehr, num_ec, output_class
        response_list = torch.stack(response_list, 0) # bs_trial, bs_ehr, num_ec, mem_dim
        return pred_list, response_list, trans_query


'''
Provide a general trainer functions for most deep learning
based methods.
'''
import math
import pdb
import os
import json
from pickletools import optimize
from typing import List, Dict, Tuple, Iterable, Type, Union, Callable, Optional
from collections.abc import Mapping
from collections import defaultdict
import warnings

import numpy as np
import torch
from torch.cuda.amp import autocast
import transformers
from torch.utils.data import DataLoader
from torch import nn
from tqdm import tqdm

from pytrial.utils.check import make_dir_if_not_exist
from pytrial.utils.parallel import DataParallelModel, unwrap_model

#TODO: accelerator (https://github.com/huggingface/accelerate) deployed for multi-gpu training.
class Trainer:
    '''
    A general trainer used to train deep learning models.

    Parameters
    ----------
    model: nn.Module
        The model to be trained.

    train_objectives: list[tuple[DataLoader, nn.Module]]
        The defined pairs of dataloaders and the loss models.

    test_data: (optional) dict or Dataset
        Depending on the implemented `get_test_dataloader` function.
        That function receives it as inputs and return test dataloader.

    test_metric: (optional) str
        Which test metric is used to judge the best checkpoint during the training.
        Only used when `test_data` is given. Should be contained in
        the returned metric dict by `evaluate` function.

    less_is_better: (optional) bool
        If the test metric is less the better.
        Ignored if no test_data and test_metric is given.

    load_best_at_end: bool
        If load the best checkpoint at the end of training.

    n_gpus: int
        How many GPUs used to kick of training.
        If set larger than 1, parallel training will be used.

    output_dir: str
        The intermediate model checkpoints during the training
        will be dump to under this dir.

    Examples
    --------
    >>> trainer = Trainer(
    ... model=model,
    ... train_objectives=[(dataloader1, loss_model1), (dataloader2, loss_model2)],
    ... )
    >>> trainer.train(
    ... epochs=10,
    ... )
    '''
    evaluated=False
    def __init__(self,
        model: nn.Module,
        train_objectives: List[Tuple[DataLoader, nn.Module]],
        test_data = None,
        test_metric = None,
        less_is_better = False,
        load_best_at_end = True,
        n_gpus = 1,
        output_dir = './checkpoints/',
        **kwargs,
        ):
        self.train_dataloader = [dataloader for dataloader,_ in train_objectives]
        self.loss_models = [loss_model for _,loss_model in train_objectives]
        self.model = model
        if test_data is not None: self.test_dataloader = self.get_test_dataloader(test_data)
        else: self.test_dataloader = None
        self.test_metric = test_metric
        self.less_is_better = less_is_better
        self.load_best_at_end = load_best_at_end
        self.output_dir = output_dir
        make_dir_if_not_exist(output_dir)
        if self.test_dataloader is not None:
            self.best_dir = os.path.join(self.output_dir, 'best')
            make_dir_if_not_exist(self.best_dir)

        # TODO add multi gpu support
        self.n_gpus = n_gpus
        if self.n_gpus > 1:
            from pytrial.utils.parallel import DataParallelModel, DataParallelCriterion
            self.model = self._wrap_model(self.model, device_ids=range(self.n_gpus))

    def train(self,
        epochs=10,
        learning_rate=2e-5,
        weight_decay=1e-4,
        warmup_ratio=0,
        scheduler='warmupcosine',
        evaluation_steps=10,
        max_grad_norm=0.5,
        use_amp=False,
        **kwargs,
        ):
        '''
        Kick of training using the provided loss model and train dataloaders.

        Parameters
        ----------
        epochs: int (default=10)
            Number of iterations (epochs) over the corpus.

        learning_rate: float (default=3e-5)
            The learning rate.

        weight_decay: float (default=1e-4)
            Weight decay applied for regularization.

        warmup_ratio: float (default=0)
            How many steps used for warmup training. 
            
            If set 0, not warmup.

        scheduler: {'constantlr','warmupconstant','warmuplinear','warmupcosine','warmupcosinewithhardrestarts'}
            Pick learning rate scheduler for warmup. Ignored if warmup_ratio <= 0.

        evaluation_steps: int (default=10)
            How many iterations 
            
            while we print the training loss and
            conduct evaluation if evaluator is given.

        max_grad_norm: float (default=0.5)
            Clip the gradient to avoid NaN.

        use_amp: bool (default=False)
            Whether or not use mixed precision training.
        '''
        self.best_score = np.inf if self.less_is_better else -np.inf

        self.score_logs = defaultdict(list)
        if use_amp:
            scaler = torch.cuda.amp.GradScaler()
        else:
            scaler = None

        num_examples = self._compute_num_train_samples(self.train_dataloader[0])
        steps_per_epoch, num_train_steps = self._build_data_iterator(epochs=epochs)
        warmup_steps = self._compute_warmup_steps(num_train_steps=num_train_steps, warmup_ratio=warmup_ratio)
        
        optimizer_param = {'lr':learning_rate, 'weight_decay':weight_decay}
        self.optimizers, self.schedulers = self._build_optimizer(scheduler=scheduler,
            loss_models=self.loss_models,
            warmup_steps=warmup_steps,
            num_train_steps=num_train_steps,
            optimizer_param=optimizer_param,
            weight_decay=weight_decay,
            )
        
        print("***** Running training *****")
        print(f"  Num examples = {num_examples}")
        print(f"  Num Epochs = {epochs}")
        print(f"  Total optimization steps = {num_train_steps}")

        train_loss_dict = defaultdict(list)
        global_step = 0
        for epoch in tqdm(range(epochs), desc='Training Epoch'):
            training_steps = 0
            self.model.train()

            for _ in tqdm(range(steps_per_epoch), desc='Iteration'):

                # train models in one iter
                self.train_one_iteration(
                        max_grad_norm=max_grad_norm,
                        warmup_steps=warmup_steps,
                        use_amp=use_amp, 
                        scaler=scaler,
                        train_loss_dict=train_loss_dict
                    )

                training_steps += 1
                global_step += 1

                if evaluation_steps>0 and global_step % evaluation_steps == 0:
                    
                    self.model.eval()
                    
                    self.evaluated = True

                    print('\n######### Train Loss {} #########'.format(global_step))
                    for key in train_loss_dict.keys():
                        print('{} {:.4f} \n'.format(key, np.mean(train_loss_dict[key])))

                    if scheduler is not None and warmup_steps > 0:
                        for scheduler_ in self.schedulers:
                            print('learning rate: {:.6f}'.format(scheduler_.get_lr()[0]))
                    
                    train_loss_dict = defaultdict(list)
                    if self.test_dataloader is not None:
                        metrics = self.evaluate()

                        print('\n######### Eval {} #########'.format(global_step))
                        for metric, metric_value in metrics.items():
                            self.score_logs[metric].append(metric_value)
                            print('{}: {:.4f}'.format(metric, metric_value))

                        if self.test_metric in metrics:
                            metric_value = metrics[self.test_metric]

                        elif self.test_metric is None:
                            warnings.warn(f'The `test_metric` was not specified, use the default metric `{metric}` as the test metric for judging early stopping.')

                        else:
                            raise ValueError(f'Do not find the specified metric `{self.test_metric}` from the evaluation results, please check either `test_metric` or the `trainer.evalute` function.')
                        
                        # update best score and save checkpoints
                        best_score = self._update_best_metric(self.best_score, metric_value, self.less_is_better)
                        if best_score != self.best_score:
                            self.score_logs['best_global_step'] = global_step
                            model = unwrap_model(self.model)
                            model.save_model(self.best_dir)
                            self.best_score = best_score
                            print(f'Best checkpoint is updated at {global_step} with {self.test_metric} {self.best_score}.')

                    else:
                        # save model checkpoint
                        output_dir = os.path.join(self.output_dir, f'./{global_step}')
                        make_dir_if_not_exist(output_dir)
                        model = unwrap_model(self.model)
                        model.save_model(output_dir)
                        print("Checkpoint saved in {} at {} steps.".format(output_dir, global_step))

        # after training
        self._save_log(self.output_dir)
        if self.load_best_at_end and self.test_dataloader is not None and self.evaluated:
            print(f'Load best ckpt from `{self.best_dir}`.')
            self.model.load_model(self.best_dir)

        print('Training completes.')

    def evaluate(self):
        '''
        Need to be created by specific tasks.

        Returns
        -------
        A dict of computed evalution metrics.
        '''
        raise NotImplementedError

    def get_test_dataloader(self, test_data):
        '''
        Need to be created by specific tasks.
        '''
        raise NotImplementedError

    def prepare_input(self, data):
        '''
        Need to be reimplemented sometimes when input data is not in the standard dict structure.
        '''
        if len(data) == 0:
            raise ValueError(
                "The batch received was empty, your model won't be able to train on it. Double-check that your "
                f"training dataset contains keys expected by the model."
            )

        if isinstance(data, Mapping):
            return type(data)({k: self.prepare_input(v) for k, v in data.items()})

        elif isinstance(data, (tuple,list)):
            return type(data)(self.prepare_input(v) for v in data)

        elif isinstance(data, torch.Tensor):
            kwargs = dict(device=self.model.device)
            return data.to(**kwargs)

        return data

    def train_one_iteration(self, 
        max_grad_norm=None,
        warmup_steps=None,
        use_amp=None, 
        scaler=None,
        train_loss_dict=None):
        '''
        Default training one iteration steps, can be subclass can reimplemented.
        '''
        skip_scheduler = False
        num_train_objectives = len(self.train_dataloader)
        for train_idx in range(num_train_objectives):
            data_iterator = self.data_iterators[train_idx]
            loss_model = self.loss_models[train_idx]
            loss_model.zero_grad()
            loss_model.train()
            optimizer = self.optimizers[train_idx]
            scheduler = self.schedulers[train_idx]

            # get a batch of data from the target data_iterator
            data = self._get_a_train_batch(data_iterator=data_iterator, train_idx=train_idx)
            data = self.prepare_input(data)

            # update model by backpropagation
            if use_amp:
                loss_value, skip_scheduler, scale_before_step = self._update_one_iteration_amp(loss_model=loss_model, data=data, optimizer=optimizer, scaler=scaler, max_grad_norm=max_grad_norm)
            else:
                loss_value = self._update_one_iteration(loss_model=loss_model, data=data, optimizer=optimizer, max_grad_norm=max_grad_norm)

            train_loss_dict[train_idx].append(loss_value.item())

            if use_amp:
                skip_scheduler = scaler.get_scale() != scale_before_step

            if not skip_scheduler and warmup_steps > 0:
                scheduler.step()

    def _build_data_iterator(self, epochs):
        self.data_iterators = [iter(dataloader) for dataloader in self.train_dataloader]
        steps_per_epoch = max([len(dataloader) for dataloader in self.train_dataloader])
        num_train_steps = int(steps_per_epoch * epochs)
        return  steps_per_epoch, num_train_steps

    def _build_optimizer(self,
        loss_models,
        warmup_steps,
        num_train_steps,
        optimizer_param,
        weight_decay,
        scheduler=None,
        ):
        optimizer_class = torch.optim.AdamW
        optimizers = []
        schedulers = []
        for loss_model in loss_models:
            print(optimizer_param)
            param_optimizer = list(loss_model.named_parameters())
            no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
            optimizer_grouped_parameters = [
                {'params': [p for n, p in param_optimizer if not any(nd in n for nd in no_decay)], 'weight_decay': weight_decay},
                {'params': [p for n, p in param_optimizer if any(nd in n for nd in no_decay)], 'weight_decay': 0.0}
            ]
            optimizer = optimizer_class(optimizer_grouped_parameters, **optimizer_param)
            if warmup_steps > 0:
                scheduler_obj = self._build_scheduler(optimizer, scheduler, warmup_steps=warmup_steps, t_total=num_train_steps)
            else:
                scheduler_obj = None
            optimizers.append(optimizer)
            schedulers.append(scheduler_obj)
        return optimizers, schedulers

    def _build_scheduler(self, optimizer, scheduler: str, warmup_steps: int, t_total: int):
        """
        Returns the correct learning rate scheduler. Available scheduler: constantlr, warmupconstant, warmuplinear, warmupcosine, warmupcosinewithhardrestarts
        """
        scheduler = scheduler.lower()
        if scheduler == 'constantlr':
            return transformers.get_constant_schedule(optimizer)
        elif scheduler == 'warmupconstant':
            return transformers.get_constant_schedule_with_warmup(optimizer, num_warmup_steps=warmup_steps)
        elif scheduler == 'warmuplinear':
            return transformers.get_linear_schedule_with_warmup(optimizer, num_warmup_steps=warmup_steps, num_training_steps=t_total)
        elif scheduler == 'warmupcosine':
            return transformers.get_cosine_schedule_with_warmup(optimizer, num_warmup_steps=warmup_steps, num_training_steps=t_total)
        elif scheduler == 'warmupcosinewithhardrestarts':
            return transformers.get_cosine_with_hard_restarts_schedule_with_warmup(optimizer, num_warmup_steps=warmup_steps, num_training_steps=t_total)
        else:
            raise ValueError("Unknown scheduler {}".format(scheduler))

    def _update_best_metric(self, best_score, metric_value, less_is_better):
        if less_is_better:
            if metric_value < best_score:
                print("New best score: from {} to {}".format(best_score, metric_value))
                return metric_value
            else:
                return best_score
        else:
            if metric_value > best_score:
                print("New best score: from {} to {}".format(best_score, metric_value))
                return metric_value
            else:
                return best_score

    def _save_log(self, output_dir):
        log_filename = os.path.join(output_dir, 'train_logs.json')
        with open(log_filename, 'w') as f:
            score_logs = str(self.score_logs)
            f.write(json.dumps(score_logs, indent=4, sort_keys=True))

    def _wrap_model(self, model):
        # wrap model to be paralleled
        model = DataParallelModel(model)
        return model

    def _compute_num_train_samples(self, dataloader):
        if isinstance(dataloader, DataLoader):
            return len(dataloader.dataset)
        else:
            # if there are more than one dataloaders for a loss_model,
            # iterate to find the last dataloader item
            return self._compute_num_train_samples(dataloader[0])

    def _compute_warmup_steps(self, num_train_steps, warmup_ratio):
        warmup_steps = math.ceil(num_train_steps * warmup_ratio) if warmup_ratio > 0 else 0
        return warmup_steps


    def _update_one_iteration(self, loss_model, data, optimizer, max_grad_norm):
        loss_model_return = loss_model(data)
        loss_value = loss_model_return['loss_value']
        loss_value.backward()
        torch.nn.utils.clip_grad_norm_(loss_model.parameters(), max_grad_norm)
        optimizer.step()
        return loss_value

    def _update_one_iteration_amp(self, loss_model, data, optimizer, scaler, max_grad_norm):
        with autocast():
            loss_return = loss_model(data)
        loss_value = loss_return['loss_value']
        scale_before_step = scaler.get_scale()
        scaler.scale(loss_value).backward()
        scaler.unscale_(optimizer)
        torch.nn.utils.clip_grad_norm_(loss_model.parameters(), max_grad_norm)
        scaler.step(optimizer)
        scaler.update()
        skip_scheduler = scaler.get_scale() != scale_before_step
        return loss_value, skip_scheduler, scale_before_step

    def _get_a_train_batch(self, data_iterator, train_idx):
        try:
            data = next(data_iterator)

        except StopIteration:
            data_iterator = iter(self.train_dataloader[train_idx])
            self.data_iterators[train_idx] = data_iterator
            data = next(data_iterator)
            
        return data



import pdb

import numpy as np
import torch
from sklearn.metrics import roc_auc_score

def rocauc(pred, label):
    '''
    For binary classification.

    Parameters
    ----------
    pred: np.ndarray or torch.Tensor
        Prediction in shape (N, ) or (N, 1)
    
    label: np.ndarray or torch.Tensor
        Prediction in shape (N, ) or (N, 1)
    '''
    if isinstance(pred, torch.Tensor): pred = pred.detach().cpu().numpy()
    if isinstance(label, torch.Tensor): label = label.cpu().numpy()
    if pred.shape != label.shape: pred = pred.reshape(label.shape)
    score = roc_auc_score(label, pred)
    return {'auc':score}

def accuracy(pred, label):
    '''
    For multiclass classification.

    Parameters
    ----------
    pred: np.ndarray or torch.Tensor
        Prediction in shape (N, C)
    
    label: np.ndarray or torch.Tensor
        Prediction in shape (N, C) or in shape (N, )
    '''
    if isinstance(pred, torch.Tensor): pred = pred.detach().cpu().numpy()
    if isinstance(label, torch.Tensor): label = label.cpu().numpy()

    if len(label.shape) == 2:
        label = np.argmax(label, 1)
    
    pred = np.argmax(pred, 1)
    score = np.mean(pred == label)
    return {'acc':score}

def macro_auc(pred, label):
    '''
    For multilabel classification.

    Parameters
    ----------
    pred: np.ndarray or torch.Tensor
        Prediction in shape (N, C)
    
    label: np.ndarray or torch.Tensor
        Prediction in shape (N, C)
    '''
    if isinstance(pred, torch.Tensor): pred = pred.detach().cpu().numpy()
    if isinstance(label, torch.Tensor): label = label.cpu().numpy()

    score = roc_auc_score(label, pred)
    return {'auc':score}

def mse(pred, label):
    '''
    For regression.

    Parameters
    ----------
    pred: np.ndarray or torch.Tensor
        Prediction in shape (N, ) or (N, 1)
    
    label: np.ndarray or torch.Tensor
        Prediction in shape (N, ) or (N, 1)
    '''
    if isinstance(pred, torch.Tensor): pred = pred.detach().cpu().numpy()
    if isinstance(label, torch.Tensor): label = label.cpu().numpy()
    if pred.shape != label.shape: pred = pred.reshape(label.shape)

    score = np.mean(np.square(pred - label))
    return {'mse': score}

METRICS_DICT = {
    'binary': rocauc,
    'multiclass': accuracy,
    'multilabel': macro_auc,
    'regression': mse,
}



'''
Some functions for mimic-iii data preprocessing.
Code is adapted from https://github.com/zzachw/PyHealth/blob/master/pyhealth/data/base_mimic.py.
'''

import numpy as np
import pandas as pd
import re
import os
import pdb

from tqdm import tqdm

import warnings
warnings.filterwarnings("ignore")

class MIMIC_Data:
    """The data template to store MIMIC data. Customized fields can be added
    in each parse_xxx methods.
    Parameters
        ----------
        patient_id
        time_duration
        selection_method
    """

    def __init__(self, patient_id, time_duration, selection_method):
        self.data = {}
        self.data['patient_id'] = str(patient_id)
        self.data['admission_list'] = []
        self.time_duration = time_duration
        self.selection_method = selection_method

    # TODO add more parse_* function on procedure and on medications

    def parse_patient(self, pd_series, mapping_dict=None):
        if mapping_dict is None:
            self.data['gender'] = pd_series['gender'].values[0]
            self.data['dob'] = pd_series['dob'].values[0]
        else:
            self.data['gender'] = pd_series[mapping_dict['gender']].values[0]
            self.data['dob'] = pd_series[mapping_dict['dob']].values[0]

    def parse_admission(self, pd_df):
        # TODO: implement the mapping dict
        for ind, row in pd_df.iterrows():
            # each admission is stored as a seperate dictionary and
            # added to admission_list
            admission_event = {}
            admission_event['admission_id'] = row['hadm_id']
            admission_event['admission_date'] = row['admittime']
            admission_event['discharge_date'] = row['dischtime']
            admission_event['death_indicator'] = int(~pd.isna(pd_df['deathtime']).values[0])

            # more elements can be added here by taking elements from row
            self.data['admission_list'].append(admission_event)

    def parse_diagnosis(self, pd_df):
        for i, admission_event in enumerate(self.data['admission_list']):
            temp_df = pd_df.loc[
                pd_df['hadm_id'] == admission_event['admission_id']]

            diag_code_list = temp_df['icd9_code'].tolist()
            admission_event['diagnosis'] = diag_code_list
            self.data['admission_list'][i] = admission_event

    def parse_prescription(self, pd_df, save_dir=''):
        if len(self.data['admission_list']) == 0:
            raise ValueError(
                "No admission information found. Parse admission info first.")

        # make saving directory if needed
        if not os.path.isdir(save_dir):
            os.mkdir(save_dir)

        for i, admission_event in enumerate(self.data['admission_list']):
            # get all the events
            temp_df = pd_df.loc[
                pd_df['hadm_id'] == admission_event['admission_id']]
            if not temp_df.empty:
                # save csv location
                # raw event
                admission_event['prescription_csv'] = str(
                    self.data['patient_id']) + '_' + str(
                    admission_event['admission_id']) + '_prescription' + '.csv'

                temp_df = temp_df.sort_values(by='startdate')

                temp_df.to_csv(
                    os.path.join(save_dir, admission_event['prescription_csv']),
                    index=False,
                )
            self.data['admission_list'][i] = admission_event

    def parse_procedure(self, pd_df):
        for i, admission_event in enumerate(self.data['admission_list']):
            temp_df = pd_df.loc[
                pd_df['hadm_id'] == admission_event['admission_id']]

            prod_code_list = temp_df['icd9_code'].tolist()
            admission_event['procedure'] = prod_code_list
            self.data['admission_list'][i] = admission_event

    def parse_icu(self, pd_df, mapping_dict=None):
        if len(self.data['admission_list']) == 0:
            raise ValueError(
                "No admission information found. Parse admission info first.")

        for i, admission_event in enumerate(self.data['admission_list']):
            # print(ind, self.data['patient_id'], admission_event['admission_id'])
            temp_df = pd_df.loc[
                pd_df['hadm_id'] == admission_event['admission_id']]
            # print(temp_df.shape)
            for ind, row in temp_df.iterrows():
                admission_event['icustay_id'] = row['icustay_id']
            self.data['admission_list'][i] = admission_event

    def parse_event(self, pd_df, save_dir='', event_mapping_df='',
                    var_list=None):

        if len(self.data['admission_list']) == 0:
            raise ValueError(
                "No admission information found. Parse admission info first.")

        # make saving directory if needed
        if not os.path.isdir(save_dir):
            os.mkdir(save_dir)

        for i, admission_event in enumerate(self.data['admission_list']):

            # get all the events
            temp_df = pd_df.loc[
                pd_df['hadm_id'] == admission_event['admission_id']]

            if not temp_df.empty:
                # save csv location
                # raw event
                admission_event['event_csv'] = str(
                    self.data['patient_id']) + '_' + str(
                    admission_event['admission_id']) + '_event' + '.csv'

                # sort it by the date and time
                temp_df = temp_df.sort_values(by='charttime')

                first_index = temp_df['charttime'].index[0]
                # print(temp_df['charttime'][first_index])
                temp_df['first_entry'] = pd.to_datetime(
                    temp_df['charttime'][first_index])

                # calculate the time difference in seconds since the first entry
                temp_df['secs_since_entry'] = (temp_df['charttime'] - temp_df[
                    'first_entry']).dt.total_seconds()
                # speed up by dropping date format
                temp_df['charttime'] = temp_df['charttime'].astype(str)
                temp_df['first_entry'] = temp_df['first_entry'].astype(str)

                # save the raw event data to local csv
                temp_df.to_csv(
                    os.path.join(save_dir, admission_event['event_csv']),
                    index=False)

                # generate episode data
                # handle all events data by aggregation during a time window
                episode_df = self.generate_episode(
                    temp_df, duration=self.time_duration,
                    event_mapping_df=event_mapping_df,
                    var_list=var_list)

                if episode_df.shape[0] >= 2:
                    # save episode location only if there are more than 2 sequencnes
                    # do not save if it is empty or just a single sequence
                    admission_event['episode_csv'] = str(
                        self.data['patient_id']) + '_' + str(
                        admission_event['admission_id']) + '_episode.csv'

                    # save the episode data to local csv
                    episode_df.to_csv(
                        os.path.join(save_dir, admission_event['episode_csv']),
                        index=False)

                # update the dictionary
                self.data['admission_list'][i] = admission_event

        return temp_df  # for debug purpose

    def write_record(self, temp_list, temp_df, var):
        # TODO: this may be sped up by aggreagte on itemid, and then use the last one
        if not temp_df.empty:
            # clean specific column
            if var == 'temperature':
                temp_df.loc[:, 'value'] = clean_temperature(temp_df).copy()
            elif var == 'diastolic blood pressure':
                temp_df.loc[:,'value'] = clean_dbp(temp_df).copy()
            elif var == 'systolic blood pressure':
                temp_df.loc[:,'value'] = clean_sbp(temp_df).copy()
            elif var == 'capillary refill rate':
                temp_df.loc[:,'value'] = clean_crr(temp_df).copy()
            elif var == 'weight':
                temp_df.loc[:,'value'] = clean_weight(temp_df).copy()
            elif var == 'height':
                temp_df.loc[:,'value'] = clean_height(temp_df).copy()
            elif var == 'fraction inspired oxygen':
                temp_df.loc[:,'value'] = clean_fio2(temp_df).copy()
            elif var == 'height':
                temp_df.loc[:,'glucose'] = clean_lab(temp_df).copy()
            elif var == 'ph':
                temp_df.loc[:,'value'] = clean_lab(temp_df).copy()
            elif var == 'oxygen saturation':
                temp_df.loc[:,'value'] = clean_o2sat(temp_df).copy()

            if self.selection_method == 'last':
                # take the last record within the range
                temp_list.append(temp_df.iloc[-1]['value'])
            elif self.selection_method == 'mean':
                temp_list.append(temp_df['value'].mean())
                # if there is no information, just skip
        else:
            temp_list.append('')

        return temp_list

    def generate_episode_headers(self, var_list):
        """Generate the header for episode file
        Parameters
        ----------
        var_list
        Returns
        -------
        """

        self.episode_headers_ = ['secs_since_entry']
        for var in var_list:
            # self.episode_headers_.extend([var, var + '_unit'])
            self.episode_headers_.append(var)

    def generate_episode(self, pd_df, duration, event_mapping_df, var_list):

        self.generate_episode_headers(var_list)

        # when it is in the memory, process it directly
        max_time_diff = pd_df['secs_since_entry'].max()

        # only keep the events we are interested in
        key_df = event_mapping_df[event_mapping_df['level2'].isin(var_list)]

        # rounding up
        n_episode = int(np.ceil(max_time_diff / duration))

        # print(n_episode)

        episode_df = pd.DataFrame(columns=self.episode_headers_)

        for j in range(n_episode):
            threshold_l = j * duration
            threshold_h = (j + 1) * duration

            # find all the events within the duration
            slice_df = pd_df.loc[(pd_df['secs_since_entry'] >= threshold_l) & (
                    pd_df['secs_since_entry'] < threshold_h)]

            # need some sort on the time so just get the last value within the range
            # can join on multiple things, weight for instance
            temp_df = key_df.merge(slice_df, left_on='itemid',
                                   right_on='itemid')

            # valid information is available, otherwise skip
            if not temp_df.empty:
                temp_df.sort_values(by='secs_since_entry', inplace=True)

                # initialize the record with timestamp
                temp_list = [threshold_h]

                # iterate over variables at interests
                for var in var_list:
                    var_df = temp_df[temp_df['level2'] == var]
                    # write records for each variable at interest
                    temp_list = self.write_record(temp_list, var_df, var)

                # append to the major episode dataframe
                epi_df = pd.DataFrame(temp_list).transpose()
                epi_df.columns = self.episode_headers_
                episode_df = pd.concat([episode_df, epi_df], axis=0)

                # ###############################################################
                # # need to iterate over rows, may not be sufficiently faster.
                # # on park
                # get the last record for each variable
                # episode_df = temp_df.groupby(['level2']).agg({
                #     'itemid' : 'last',
                #     'value' : 'last',
                #     'valueuom' : 'last',
                #     })
                # self.a = episode_df
                # print(episode_df)
                # ###############################################################

        # return the episode df
        return episode_df


##############################################################################

def parallel_parse_tables(patient_id_list, patient_df, admission_df, icu_df,
                          event_df, event_mapping_df,
                          prescription_df, procedure_df, diagnosis_df,
                          duration,
                          selection_method, var_list, save_dir):
    """Parallel methods to process patient information in batches

    Parameters
    ----------
    patient_id_list: list[int]
        A list of patient subject_id.

    patient_df: pd.DataFrame
        Patient table read from 'PATIENTS.csv'.

    admission_df: pd.DataFrame
        Admission table read from 'ADMISSIONS.csv'.

    icu_df: pd.DataFrame
        ICU stay table read from 'ICUSTAYS.csv'.

    event_df: pd.DataFrame
        Patient events read fom 'CHARTEVENTS.csv' and 'OUTPUTEVENTS.csv'.

    prescription_df: pd.DataFrame
        Prescription table read from 'PRESCRIPTIONS.csv'.

    procedure_df: pd.DataFrame
        Procedure table read from 'PROCEDURES_ICD.csv'.

    diagnosis_df: pd.DataFrame
        Diagnosis table read from 'DIAGNOSES_ICD.csv'.

    var_list: list[str]
        Selected event types for event_df.

    Returns
    -------
    """
    valid_data_list, valid_id_list = [], []

    # for i in tqdm(range(len(patient_id_list))):
    for i in tqdm(range(len(patient_id_list))):
        p_id = patient_id_list[i]
        # print('Processing Patient', i + 1, p_id)
        # initialize the
        temp_data = MIMIC_Data(p_id, duration, selection_method)
        p_df = patient_df.loc[patient_df['subject_id'] == p_id]
        a_df = admission_df.loc[admission_df['subject_id'] == p_id]
        i_df = icu_df.loc[icu_df['subject_id'] == p_id]
        e_df = event_df.loc[event_df['subject_id'] == p_id]
        drug_df = prescription_df.loc[prescription_df['subject_id']==p_id]
        diag_df = diagnosis_df.loc[diagnosis_df['subject_id']==p_id]
        prod_df = procedure_df.loc[procedure_df['subject_id']==p_id]

        if not p_df.empty:
            if p_df.shape[0] > 1:
                raise ValueError("Patient ID cannot be repeated")
            temp_data.parse_patient(p_df)

        if not a_df.empty:
            temp_data.parse_admission(a_df)

        if not i_df.empty:
            temp_data.parse_icu(i_df)

        if not e_df.empty:
            temp_data.parse_event(e_df, save_dir=save_dir,
                                  event_mapping_df=event_mapping_df,
                                  var_list=var_list)

        if not diag_df.empty:
            temp_data.parse_diagnosis(diag_df)

        if not drug_df.empty:
            temp_data.parse_prescription(drug_df, save_dir=save_dir)

        if not prod_df.empty:
            temp_data.parse_procedure(prod_df)


        '''
        `temp_data.data` is a dict
        dict{
        'patient_id':xxx,
        'admission_list': [
            {
                'admission_id':xxx, 'admission_data':xxxx,
                'event_csv': 'xxx.csv',
                'episode_csv':'xxxx.csv',
                'diagnosis': [icd1,icd2,...],
                'procedure': [icd1,icd2,...],
                'prescription_csv': 'xxx.csv',
                ....
            },
            ],
        'gender':'F',
        'dob': '2022-02-02 00:00:00',
        ...
        }
        '''

        valid_data_list.append(temp_data)
        valid_id_list.append(p_id)

    return valid_data_list, valid_id_list


# Temperature: map Farenheit to Celsius, some ambiguous 50<x<80
def clean_temperature(df):
    try:
        v = df.value.astype('float32').copy()
    except ValueError:
        print("could not convert string to float in clean_temperature. "
              "Set to NaN")
        v = df.value
        v[:] = np.nan
        return v
    else:
        idx = df.valueuom.fillna('').apply(lambda s: 'F' in s.lower()) | (
                v >= 79)
        v.loc[idx] = (v[idx] - 32) * 5. / 9
        return v

def clean_sbp(df):
    v = df.value.astype(str).copy()
    idx = v.apply(lambda s: '/' in s)
    v.loc[idx] = v[idx].apply(lambda s: re.match('^(\d+)/(\d+)$', s).group(1))
    return v.astype(float)


def clean_dbp(df):
    v = df.value.astype(str).copy()
    idx = v.apply(lambda s: '/' in s)
    v.loc[idx] = v[idx].apply(lambda s: re.match('^(\d+)/(\d+)$', s).group(2))
    return v.astype(float)


def clean_crr(df):
    v = pd.Series(np.zeros(df.shape[0]), index=df.index)
    v[:] = np.nan

    # when df.value is empty, dtype can be float and comparision with string
    # raises an exception, to fix this we change dtype to str
    df_value_str = df.value.astype(str)

    v.loc[(df_value_str == 'Normal <3 secs') | (df_value_str == 'Brisk')] = 0
    v.loc[
        (df_value_str == 'Abnormal >3 secs') | (df_value_str == 'Delayed')] = 1
    return v


# Weight: some really light/heavy adults: <50 lb, >450 lb, ambiguous oz/lb
# Children are tough for height, weight
def clean_weight(df):
    try:
        v = df.value.astype('float32').copy()
    except ValueError:
        print("could not convert string to float in clean_weight. Set to NaN")
        v = df.value
        v[:] = np.nan
        return v
    else:
        # ounces
        idx = df.valueuom.fillna('').apply(lambda s: 'oz' in s.lower())
        v.loc[idx] = v[idx] / 16.
        # pounds
        idx = idx | df.valueuom.fillna('').apply(lambda s: 'lb' in s.lower())
        v.loc[idx] = v[idx] * 0.453592
        return v


# Height: some really short/tall adults: <2 ft, >7 ft)
# Children are tough for height, weight
def clean_height(df):
    try:
        v = df.value.astype('float32').copy()
    except ValueError:
        print("could not convert string to float in clean_height. Set to NaN")
        v = df.value
        v[:] = np.nan
        return v
    else:
        idx = df.valueuom.fillna('').apply(lambda s: 'in' in s.lower())
        v.loc[idx] = np.round(v[idx] * 2.54)
        return v


# FIO2: many 0s, some 0<x<0.2 or 1<x<20
def clean_fio2(df):
    try:
        v = df.value.astype('float32').copy()
    except ValueError:
        print("could not convert string to float in clean_fio2. Set to NaN")
        v = df.value
        v[:] = np.nan
        return v
    else:
        ''' The line below is the correct way of doing the cleaning,
        since we will not compare 'str' to 'float'.
        If we use that line it will create mismatches from the data of the
        paper in ~50 ICU stays. The next releases of the benchmark should use this line.
        '''
        # idx = df.valueuom.fillna('').apply(lambda s: 'torr' not in s.lower()) & (v>1.0)

        ''' The line below was used to create the benchmark dataset that the
        paper used. Note this line will not work in python 3,
        since it may try to compare 'str' to 'float'.
        '''
        # idx = df.valueuom.fillna('').apply(lambda s: 'torr' not in s.lower()) & (df.value > 1.0)

        ''' The two following lines implement the code that was used to create
        the benchmark dataset that the paper used.
        This works with both python 2 and python 3.
        '''
        is_str = np.array(map(lambda x: type(x) == str, list(df.value)),
                          dtype=np.bool)
        idx = df.valueuom.fillna('').apply(
            lambda s: 'torr' not in s.lower()) & (
                      is_str | (~is_str & (v > 1.0)))

        v.loc[idx] = v[idx] / 100.
        return v


# GLUCOSE, PH: sometimes have ERROR as value
def clean_lab(df):
    v = df.value.copy()
    idx = v.apply(
        lambda s: type(s) is str and not re.match('^(\d+(\.\d*)?|\.\d+)$', s))
    v.loc[idx] = np.nan
    return v.astype(float)


# O2SAT: small number of 0<x<=1 that should be mapped to 0-100 scale
def clean_o2sat(df):
    # change "ERROR" to NaN
    v = df.value.copy()
    idx = v.apply(
        lambda s: type(s) is str and not re.match('^(\d+(\.\d*)?|\.\d+)$', s))
    v.loc[idx] = np.nan

    v = v.astype(float)
    idx = (v <= 1)
    v.loc[idx] = v[idx] * 100.
    return v

def convert_to_3digit_icd9(dxStr):
	if dxStr.startswith('E'):
		if len(dxStr) > 4: return dxStr[:4]
		else: return dxStr
	else:
		if len(dxStr) > 3: return dxStr[:3]
		else: return dxStr



import os
import glob
import pdb
import subprocess
import sys

def check_model_dir(experiment_id, root_dir='./experiments_records'):
    """
    Check whether the checkouts/results folders of current experiment(exp_id) exist,
        If not, will create both folders
    Parameters
    ----------
    experiment_id : str, optional (default='init.test')
        name of current experiment

    root_dir :  str,
        root dir of current project
    """
    if os.path.exists(root_dir) is False:
        os.mkdir(root_dir)
    exp_root = os.path.join(root_dir, experiment_id)
    if os.path.exists(exp_root) is False:
        os.mkdir(exp_root)
    checkout_dir = os.path.join(exp_root, 'checkpoints')
    result_dir = os.path.join(exp_root, 'results')
    if os.path.exists(checkout_dir) is False:
        os.mkdir(checkout_dir)
    if os.path.exists(result_dir) is False:
        os.mkdir(result_dir)

def check_checkpoint_file(input_dir, suffix='pth.tar'):
    '''
    Check whether the `input_path` is directory or to the checkpoint file.
        If it is a directory, find the only 'pth.tar' file under it.

    Parameters
    ----------
    input_path: str
        The input path to the pretrained model.

    suffix: 'pth.tar' or 'model'
        The checkpoint file suffix;
        If 'pth.tar', the saved model is a torch model.
        If 'model', the saved model is a scikit-learn based model.
    '''
    suffix = '.' + suffix
    if input_dir.endswith(suffix):
        return input_dir

    ckpt_list = glob.glob(os.path.join(input_dir, '*'+suffix))
    assert len(ckpt_list) <= 1, f'Find more than one checkpoints under the dir {input_dir}, please specify the one to load.'
    assert len(ckpt_list) > 0, f'Do not find any checkpoint under the dir {input_dir}.'
    return ckpt_list[0]

def check_model_config_file(input_dir):
    '''
    Check whether the `input_path` is directory or to the `model_config.json` file.
        If it is a directory, find the only '.json' file under it.

    Parameters
    ----------
    input_path: str
        The input path to the pretrained model.

    '''
    if input_dir.endswith('.json'):
        return input_dir

    if not os.path.isdir(input_dir):
        # if the input_dir is the given checkpoint model path,
        # we need to find the config file under the same dir.
        input_dir = os.path.dirname(input_dir)

    ckpt_list = glob.glob(os.path.join(input_dir, '*.json'))

    if len(ckpt_list) == 0:
        return None

    # find model_config.json under this input_dir
    model_config_name = [config for config in ckpt_list if 'model_config.json' in config]
    if len(model_config_name) == 1:
        return model_config_name[0]

    # if no model_config.json found, retrieve the only .json file.
    assert len(ckpt_list) <= 1, f'Find more than one config .json under the dir {input_dir}.'
    return ckpt_list[0]

def make_dir_if_not_exist(path):
    if not os.path.exists(path):
        os.makedirs(path)

def install_package(package):
    subprocess.check_call([sys.executable, "-m", "pip", "install", package])



import torch
import numpy as np
import sklearn.metrics
			
def rel_err(score, label, K):
	avg_delta=torch.sum(torch.gather(label, 1, torch.argsort(score, axis=1, descending=True)[:,:K]), dim=1)
	max_delta=torch.sum(torch.sort(label, descending=True, axis=1)[0][:,:K], dim=1)
	return ((max_delta - avg_delta)/max_delta).tolist()

def recall_K(score, label, K):
	model_ind = torch.argsort(score, axis=1, descending=True)[:,:K].detach().cpu().tolist()
	model_ind = [[j for j in model_ind[i] if label[i][j] != 0] for i in range(len(score))] # number of rel. items in the model's top-K

	gt_ind = torch.argsort(label, axis=1, descending=True)[:,:K].detach().cpu().tolist()
	gt_ind = [[j for j in gt_ind[i] if label[i][j] != 0] for i in range(len(label))] # number of actual rel. items

	if gt_ind:
		return [len(set(m_i)&set(g_i))/len(g_i) for m_i, g_i in zip(model_ind, gt_ind) if g_i]
	else:
		return []

def precision_K(score, label, K):
	model_ind = torch.argsort(score, axis=1, descending=True)[:,:K].detach().cpu().tolist()
	model_ind = [[j for j in model_ind[i] if label[i][j] != 0] for i in range(len(score))] # number of rel. items in the model's top-K

	gt_ind = torch.argsort(label, axis=1, descending=True)[:,:K].detach().cpu().tolist()
	gt_ind = [[j for j in gt_ind[i] if label[i][j] != 0] for i in range(len(label))] # number of actual rel. items

	if gt_ind:
		return [len(set(m_i)&set(g_i))/len(m_i) for m_i, g_i in zip(model_ind, gt_ind) if m_i]
	else:
		return []

def NDCG_K(score, label, K):
	nz_scores = [[score[i][j].item() for j in range(len(label[i]))] for i in range(len(score))]
	nz_labels = [[label[i][j].item() for j in range(len(label[i]))] for i in range(len(label))]
	
	if nz_labels:
		if len(nz_labels[0]) >= K:
			return [sklearn.metrics.ndcg_score(np.asarray([nz_labels[i]]), np.asarray([nz_scores[i]]), k=K) for i in range(len(score))]
		else:
			try:
				return [sklearn.metrics.ndcg_score(np.asarray([nz_labels[i]]), np.asarray([nz_scores[i]])) for i in range(len(score))]
			except: 
				return []

def compute_repn(score, label, K, eth_sel):
	net_inds = torch.argsort(score, axis=1, descending=True)[:,:K].squeeze(-1)
	base_inds = torch.argsort(label, axis=1, descending=True)[:,:K].squeeze(-1)
	net_weights = torch.gather(label, 1, net_inds)
	base_weights = torch.gather(label, 1, base_inds)
	net_weights = net_weights/torch.sum(net_weights, 1, keepdim=True)
	base_weights = base_weights/torch.sum(base_weights, 1, keepdim=True)
	net_weights[net_weights != net_weights] = 1 / K # Handle division by zero and set to even weighting
	base_weights[base_weights != base_weights] = 1 / K # Handle division by zero and set to even weighting
	net_eth = torch.gather(eth_sel, 1, net_inds.unsqueeze(-1).repeat(1,1,eth_sel.size(-1))) * net_weights.unsqueeze(-1) / 100
	base_eth = torch.gather(eth_sel, 1, base_inds.unsqueeze(-1).repeat(1,1,eth_sel.size(-1))) * base_weights.unsqueeze(-1) / 100
	net_repn = torch.sum(net_eth, dim=1)
	base_repn = torch.sum(base_eth, dim=1)
	return net_repn, base_repn


import pdb

import torch
from torch import nn
import torch.nn.functional as F
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import DataLoader
import numpy as np


from pytrial.utils.check import (
    check_checkpoint_file, check_model_dir, check_model_config_file, make_dir_if_not_exist
)
from .base import PatientTrialMatchBase
from ..data import TrialCollator, PatientCollator
from ..losses import DeepEnrollLoss
from ..trainer import PatientTrialTrainer


class DeepEnroll(PatientTrialMatchBase):
    '''
    Leverage DeepEnroll model for patient-trial matching [1]_.

    One patient's label = [[0,1,2,3], [2,3,4,5]] where
    the first is a list of indices of inclusion criteria that the patient satisfies.
    the second is a list of indices of exclusion criteria that the patient does not satisfy.
    for the other criteria we dont know if the patient satisfied or not.

    In this regard, we need to:
    
    (1) predict "match" for all inclusion criteria of a trial for recruiting the patient
    
    (2) predict "unmatch" for all exclusion criteria of a trial for recruiting the patient

    Note that the dimension of the output logits indicates:
    
    - 0 is unmatch
    - 1 is match
    - 2 is unknown

    Parameters
    ----------
    order: list[str]
        The order of events within each visit in the patient's EHR data, e.g., orders=['diag','med','prod'].

    vocab_size: int
        The vocabular size of each patient's EHR event types, e.g., diag, med, prod.

    max_visit: int
        Maximum number of visits to load for EHRs inputs.

    word_dim: int
        The dimension of input word embeddings for encoding eligibility criteria.

    conv_dim: int
        The dimension of convolutional layers for processing eligibility criteria embeddings.

    mem_dim: int
        The hidden dimension of the EHR memory network (encode patient EHRs).

    mlp_dim: int
        The hidden dimension of the MLP layers in the Query Network.

    demo_dim: int
        The input dimensions of patient demographic information, e.g., age, gender.

    margin: float
        The margin when compute nn.CosineEmbeddingLoss on patient embedding and inclusion/exclusion criteria embeddings.
        Refer to Eq. (12) of the reference paper.
    
    epochs: int, optional (default=50)
        Number of iterations (epochs) over the training data.

    batch_size: int, optional (default=512)
        Number of samples in each training batch.

    learning_rate: float, optional (default=1e-3)
        The learning rate.
    
    weight_decay: float (default=0)
        The weight decay during training.

    num_worker: int
        Number of workers used to do dataloading during training.

    device: str
        The model device.
    
    Notes
    -----
    .. [1] Zhang, Xingyao, et al. "DeepEnroll: patient-trial matching with deep embedding and entailment prediction." Proceedings of The Web Conference 2020. 

    '''
    def __init__(self,
        order,
        vocab_size,
        max_visit=20,
        word_dim=768,
        conv_dim=128,
        mem_dim=320,
        mlp_dim=512,
        demo_dim=3,
        margin=0.3,
        batch_size=512,
        epochs=50,
        learning_rate=1e-3,
        weight_decay=0,
        num_worker=0,
        device='cpu',
        experiment_id='test',
        ) -> None:
        super().__init__(experiment_id)
        self.config = {
            'order':order,
            'max_visit':max_visit,
            'vocab_size':vocab_size,
            'word_dim':word_dim,
            'conv_dim':conv_dim,
            'mem_dim':mem_dim,
            'mlp_dim':mlp_dim,
            'demo_dim':demo_dim,
            'margin':margin,
            'batch_size':batch_size,
            'learning_rate':learning_rate,
            'weight_decay':weight_decay,
            'epochs':epochs,
            'num_worker':num_worker,
            'device':device,
            }
        self.device = device
        self.config['total_vocab_size'] = sum(vocab_size)
        self._build_model()
    
    def predict(self, test_data, return_dict=False):
        '''
        Predict the matching of patient and ECs of each target trial.
        Will make predictions for each patient, go through all included trials.
        For example, given 1000 patients and 10 trials, will generate predictions
        as the shape of ``1000 x 10 x num_eligibility_criteria``.

        Parameters
        ----------
        test_data: dict
            A dict contains patient and trial data, respecitvely. 
            
            `test_data` = 
            
            {
            
            'patient': `pytrial.tasks.trial_patient_match.data.PatientData`, 
            
            'trial': `pytrial.tasks.trial_patient_match.data.TrialData`
            
            }
        
        return_dict: bool
            If return results stored in dictionary, otherwise return tuples of results

        '''
        self._input_data_check(test_data)
        patient_dataloader, trial_dataloader = self.get_test_dataloader(test_data)
        pred_res = self._predict_on_dataloader(patient_dataloader, trial_dataloader)

        if return_dict:
            pred_res['pred_trial'] = self._tuple_result_to_dict(pred_res['pred_trial'])

        return pred_res

    def fit(self, train_data, valid_data=None):
        '''
        Fit patient-trial matching model.

        Parameters
        ----------
        train_data: dict
            A dict contains patient and trial data.
            
            `train_data` = 
            
            {

            'patient': `pytrial.tasks.trial_patient_match.data.PatientData`,
            
            'trial': `pytrial.tasks.trial_patient_match.data.TrialData`,

            }

        valid_data: dict
            A dict contains patient and trial data for evaluation. Same format as
            `train_data`.

        '''
        self._input_data_check(train_data)
        if valid_data is not None: self._input_data_check(valid_data)
        self._fit_model(train_data, valid_data)

    def save_model(self, output_dir):
        '''
        Save the learned patient-match model to the disk.

        Parameters
        ----------
        output_dir: str or None
            The dir to save the learned model.
            If set None, will save model to `self.checkout_dir`.
        '''
        if output_dir is not None:
            make_dir_if_not_exist(output_dir)
        else:
            output_dir = self.checkout_dir
        self._save_config(self.config, output_dir=output_dir)
        self._save_checkpoint({'model':self.model}, output_dir=output_dir)


    def load_model(self, checkpoint):
        '''
        Load model and the pre-encoded trial embeddings from the given
        checkpoint dir.

        Parameters
        ----------
        checkpoint: str
            The input dir that stores the pretrained model.
            
            - If a directory, the only checkpoint file `*.pth.tar` will be loaded.
            - If a filepath, will load from this file.
        '''
        checkpoint_filename = check_checkpoint_file(checkpoint)
        config_filename = check_model_config_file(checkpoint)
        state_dict = torch.load(checkpoint_filename)
        if config_filename is not None:
            config = self._load_config(config_filename)
            self.config.update(config)
        self.model = state_dict['model']


    def get_train_dataloader(self, train_data):
        trial_data = train_data['trial']
        patient_data = train_data['patient']

        patient_dataloader = DataLoader(patient_data,
            batch_size=self.config['batch_size'],
            num_workers=self.config['num_worker'],
            pin_memory=True,
            shuffle=True,
            collate_fn=PatientCollator(
                config={
                    'visit_mode':train_data['patient'].metadata['visit']['mode'],
                    'label_mode':train_data['patient'].metadata['label']['mode'],
                    }
                ),
            )
        
        trial_dataloader = DataLoader(trial_data,
            batch_size=self.config['batch_size'],
            num_workers=self.config['num_worker'],
            pin_memory=True,
            shuffle=True,
            collate_fn=TrialCollator(),
        )

        return patient_dataloader, trial_dataloader


    def get_test_dataloader(self, test_data):
        trial_data = test_data['trial']
        patient_data = test_data['patient']

        patient_dataloader = DataLoader(patient_data,
            batch_size=self.config['batch_size'],
            num_workers=self.config['num_worker'],
            pin_memory=False,
            shuffle=False,
            collate_fn=PatientCollator(
                config={
                    'visit_mode':patient_data.metadata['visit']['mode'],
                    'label_mode':patient_data.metadata['label']['mode'],
                    }
                ),
            )
        
        trial_dataloader = DataLoader(trial_data,
            batch_size=1, # get one trial once when making predictions
            num_workers=self.config['num_worker'],
            pin_memory=False,
            shuffle=False,
            collate_fn=TrialCollator(),
        )

        return patient_dataloader, trial_dataloader

    def _build_model(self):
        self.model = BuildModel(
            total_vocab_size=self.config['total_vocab_size'],
            word_dim=self.config['word_dim'],
            conv_dim=self.config['conv_dim'],
            mem_dim=self.config['mem_dim'],
            demo_dim=self.config['demo_dim'],
            mlp_dim=self.config['mlp_dim'],
        )
        if 'cuda' in self.device:
            self.model.cuda()

    def _fit_model(self, train_data, valid_data):
        patient_dataloader, trial_dataloader = self.get_train_dataloader(train_data)
        loss_models = self._build_loss_model()
        # need to customize the training process in trainer
        train_objectives = [[[patient_dataloader,trial_dataloader], loss_model] for loss_model in loss_models]
        trainer = PatientTrialTrainer(model=self,
            train_objectives=train_objectives,
            test_data=valid_data,        
        )
        trainer.train(**self.config)
    
    @torch.no_grad()
    def _predict_on_dataloader(self, patient_dataloader, trial_dataloader, return_label=False):
        '''Return trial level matching predictions.
        Return label for evaluation if required.
        '''
        self.eval()

        # record predictions
        pred_all, label_all = [], []
        pred_inc, pred_exc = [], []

        # record labels
        label_trial_inc, label_trial_exc = [], []
        label_patient_inc, label_patient_exc = [], []

        for i, batch_trial in enumerate(trial_dataloader):
            
            pred_patient_inc, pred_patient_exc = [], []

            label_trial = []
            label_trial_inc_, label_trial_exc_ = [], []

            for batch_ehr in patient_dataloader:
                inputs = {}
                inputs.update(batch_ehr)
                inputs.update(batch_trial)
                inputs = self._prepare_input(inputs)
                pred = self.model(inputs)
                pred_patient_inc.append(pred['logit_inc'])
                pred_patient_exc.append(pred['logit_exc'])

                if return_label:
                    # match trial ec idnex with patient index
                    match_inc, match_inc_ec = self._match_patient_with_trial_ec(batch_trial=batch_trial, batch_ehr=batch_ehr, ec='inc', ec_idx=0)
                    match_exc, match_exc_ec = self._match_patient_with_trial_ec(batch_trial=batch_trial, batch_ehr=batch_ehr, ec='exc', ec_idx=1)
                    match_label = match_inc * ~match_exc
                    label_trial.append(match_label.astype(int))
                    label_trial_inc_.append(match_inc_ec) # bs_patient, num_ec
                    label_trial_exc_.append(match_exc_ec)

                    if i == 0:
                        # get patient label
                        [label_patient_inc.append(y[0]) for y in batch_ehr['y']]
                        [label_patient_exc.append(y[1]) for y in batch_ehr['y']]


            if return_label:
                label_trial_inc_ = np.concatenate(label_trial_inc_)
                label_trial_exc_ = np.concatenate(label_trial_exc_)
                label_trial_inc.append(label_trial_inc_)
                label_trial_exc.append(label_trial_exc_)
                label_all.append(np.concatenate(label_trial))

            # get a batch of trial to all EHRs matching
            pred_patient_inc = torch.cat(pred_patient_inc, 1) # 1, num_patients, num_inc_ec, output_class
            pred_patient_exc = torch.cat(pred_patient_exc, 1) # 1, num_patients, num_exc_ec, output_class

            pred_trial_inc = self._match_trial_for_patients(pred_patient_inc, 'inc')
            pred_trial_exc = self._match_trial_for_patients(pred_patient_exc, 'exc')
            pred_trial = pred_trial_exc * pred_trial_inc # iff both inc and exc are matched, shape=(bs_trial, bs_patient)

            pred_all.append((batch_trial['nct_id'], pred_trial.detach().cpu().numpy()))
            pred_inc.append(pred_patient_inc.detach().cpu().numpy())
            pred_exc.append(pred_patient_exc.detach().cpu().numpy())        
        
        return_res = {
            'pred_trial':pred_all, # trial-level prediction
            'pred_inc': pred_inc, # inclusion-level prediction
            'pred_exc': pred_exc, # exclusion-level prediction
        }

        if return_label:
            return_res['label_trial'] = np.stack(label_all) # trial-level labels
            return_res['label_patient_inc'] = label_patient_inc # patient inclusion criteria label
            return_res['label_patient_exc'] = label_patient_exc # patient exclusion criteria label
            return_res['label_trial_inc'] = label_trial_inc # trial inclusion criteria index
            return_res['label_trial_exc'] = label_trial_exc # trial exclusion criteria index

        return return_res

    def _build_loss_model(self):
        return [DeepEnrollLoss(self.model)]

    def _prepare_input(self, data):
        '''
        Prepare inputs for DeepEnroll model encoding.
        {
            'v': patient visit events,
            'x': patient features,
            'y': which ECs patient are matched,
            'inc_ec_index': sampled batch of inclusion criteria index,
            'exc_ec_index': sampled batch of exclusion criteria index,
            'inc_ec_index': sampled batch of inclusion criteria embedding,
            'exc_ec_index': sampled batch of exclusion criteria embedding,
        }
        '''
        inputs = self._prepare_ehr_input(data=data)
        trial_inputs = self._prepare_trial_patient_match_input(data=data)
        inputs.update(trial_inputs)
        return inputs

    def _prepare_ehr_input(self, data):
        inputs = {}
        visits = data['v']
        feature = data['x']

        if not isinstance(feature, torch.Tensor): feature = torch.tensor(feature)
        feature = feature.to(self.device)
        v_lengths = [len(visits[self.config['order'][0]][idx][:self.config['max_visit']]) for idx in range(len(visits[self.config['order'][0]]))]
        mask = torch.zeros(len(v_lengths), max(v_lengths))

        v = torch.zeros(len(v_lengths), max(v_lengths), self.config['total_vocab_size'])
        for idx in range(len(v_lengths)):
            v[idx,:v_lengths[idx]] = torch.tensor(self._translate_dense_visits_to_sparse({k: visits[k][idx][:self.config['max_visit']] for k in visits}))
            mask[idx,:v_lengths[idx]] = 1

        if 'cuda' in self.device:
            v = v.cuda()
            mask = mask.cuda()
            feature = feature.cuda()

        inputs['x'] = feature.float()
        inputs['mask'] = mask
        inputs['v'] = v
        return inputs

    def _prepare_trial_patient_match_input(self, data):
        outputs = {}
        inc_inputs = self._prepare_ec_input(data, 'inc', 0) # patient label 0 corresponds to inclusion
        exc_inputs = self._prepare_ec_input(data, 'exc', 1) # patient label 1 corresponds to exclusion
        outputs.update(inc_inputs)
        outputs.update(exc_inputs)
        return outputs

    def _prepare_ec_input(self, data, ec, label_idx):
        patient_label = data['y']
        inc_label = []
        inc_idx = data[f'{ec}_ec_index']
        inc_emb = data[f'{ec}_ec_emb']
        max_n_inc = [len(l) for l in inc_idx]
        inc_mask = torch.zeros(len(max_n_inc), max(max_n_inc))

        num_patient = len(patient_label)
        num_sample = min(len(max_n_inc), num_patient)

        inc_emb_ts = []
        for idx in range(num_sample): 
            # go through all trials
            # if len(patients) > len(trials)
            # only take the top len(trials) samples
            ec_embs = torch.zeros(max(max_n_inc), inc_emb[0].shape[-1])
            ec_embs[:max_n_inc[idx]] = inc_emb[idx]
            inc_emb_ts.append(ec_embs)
            inc_mask[idx][:max_n_inc[idx]] = 1
            p_label = patient_label[idx][label_idx]
            
            # each trial match each patient
            # trial1 -> patient1
            # trial2 -> patient2
            # ...
            # trial10 -> patient10
            # the remaining of patients will be ignored if len(trial) == 10
            inc_label_ = np.in1d(inc_idx[idx], p_label)

            if ec == 'inc': # inclusion
                inc_label_ = torch.tensor(inc_label_.astype(int)) # match is 1
                inc_label_[inc_label_==0]  = 2 # for the other inclusion it is unknown
            else: # exclusion
                inc_label_ = torch.tensor((~inc_label_).astype(int)) # unmatch is 0
                inc_label_[inc_label_!=0] = 2 # for the other exclusion it is unknown

            if 'cuda' in self.device: inc_label_ = inc_label_.cuda()
            inc_label.append(inc_label_)

        inc_emb_ts = torch.stack(inc_emb_ts)

        if 'cuda' in self.device:
            inc_emb_ts = inc_emb_ts.cuda()
            inc_mask = inc_mask.cuda()

        inc_label = pad_sequence(inc_label, batch_first=True)

        outputs = {
            f'{ec}_emb': inc_emb_ts,
            f'{ec}_emb_mask': inc_mask,
            f'{ec}_label': inc_label,
            }
        return outputs
    
    def _match_patient_with_trial_ec(self, batch_trial, batch_ehr, ec, ec_idx):
        trial_inc_idx = batch_trial[f'{ec}_ec_index']# assume test dataloader only has 1 sample each batch
        patient_label = batch_ehr['y']
        match_label, match_ec_label = [], []
        for p_label in patient_label:
            matched = np.in1d(trial_inc_idx, p_label[ec_idx]).astype(int)
            
            if ec_idx == 0: # inclusion
                label = sum(matched) == len(trial_inc_idx)
                matched[matched==0] = 2 # unk
            else: # exclusion
                label = sum(matched) == 0
                matched[matched==0] = 2 # unk
                matched[matched==1] = 0 # unmatch
            
            match_ec_label.append(matched)
            match_label.append(label)

        return np.array(match_label), np.stack(match_ec_label)

class BuildModel(nn.Module):
    '''
    Build DeepEnroll model.
    '''
    def __init__(self, 
        total_vocab_size,
        word_dim, 
        conv_dim,
        mem_dim,
        demo_dim,
        mlp_dim,
        ) -> None:
        super().__init__()
        self.ec_network = ECEmbedding(word_dim=word_dim, conv_dim=conv_dim)
        self.ehr_network = EHRNetwork(total_vocab_size=total_vocab_size, word_dim=word_dim, mem_dim=mem_dim, demo_dim=demo_dim)
        self.query_network = Alignment(mem_dim=mem_dim, conv_dim=conv_dim, mlp_dim=mlp_dim)

    def forward(self, inputs):
        ehr = inputs['v'] # tensor bs, n_visit, n_event
        demo = inputs['x'] # tensor bs, n_feature
        ehr_mask = inputs['mask']

        memory = self.ehr_network(ehr, demo, ehr_mask) # batch_size, 2, mem_dim


        inc_ec, inc_ec_mask = inputs['inc_emb'], inputs['inc_emb_mask']
        inc_emb, inc_ec_emb = self.ec_network(inc_ec, inc_ec_mask) # bs, num_ec, 512

        exc_ec, exc_ec_mask = inputs['exc_emb'], inputs['exc_emb_mask']
        exc_emb, exc_ec_emb = self.ec_network(exc_ec, exc_ec_mask) # bs, num_ec, 512

        # print('BuildModel ehr', ehr.shape, memory.shape, inc_ec.shape, exc_emb.shape)


        ouptut_inc, response_inc, query_inc = self.query_network(memory, inc_ec_emb, inc_ec_mask) # bs_trial, bs_ehr, num_ec, 2
        output_exc, response_exc, query_exc = self.query_network(memory, exc_ec_emb, exc_ec_mask) # bs_trial, bs_ehr, num_ec, 2

        # pred_inc = torch.softmax(ouptut_inc, dim=-1)
        # pred_exc = torch.softmax(output_exc, dim=-1)
        return {'logit_inc':ouptut_inc, 
                'logit_exc':output_exc, 
                'response_inc':response_inc,
                'response_exc':response_exc,
                'query_inc':query_inc,
                'query_exc':query_exc,
                }


class ECEmbedding(nn.Module):
    def __init__(self, word_dim, conv_dim):
        super(ECEmbedding, self).__init__()
        ## word dim 768
        ## conv_dim 128  128*4 is output dim
        self.mlp = nn.Linear(word_dim, conv_dim * 4)

    def forward(self, input, mask):
        #Input: B * L * embd_dim
        #Mask: B * L

        # print(input.shape, 'ECEmbedding')
        h = torch.sum(input * torch.tensor(mask, device=input.device, dtype=torch.float32).unsqueeze(-1), 1)
        o1 = torch.relu(self.mlp(h))
        B,L,dim = input.shape 
        h2 = input.view(-1,dim)
        o2 = torch.relu(self.mlp(h2))
        o2 = o2.view(B,L,-1)
        # print(o1.shape, o2.shape, 'ECEmbedding')
        return o1, o2


class EHRNetwork(nn.Module):
    def __init__(self, total_vocab_size, word_dim, mem_dim, demo_dim):
        super(EHRNetwork, self).__init__()
        self.mem_dim = mem_dim
        self.ehr_embedding_matrix = nn.Linear(total_vocab_size, word_dim, bias=False) 
        ## total_vocab_size=3500 
        ## mem_dim 320 
        self.mlp = nn.Linear(word_dim, mem_dim)


    def forward(self, input, demo, mask):
        # print(input.shape, 'EHRNetwork1')
        h = self.ehr_embedding_matrix(input)  ### 128,14,3500 -> 128,14,word_dim 
        h = torch.sum(h * mask.unsqueeze(-1), dim=1) ### 128,word_dim 
        o = torch.relu(self.mlp(h)) ### 128,320 
        # print(o.shape, 'EHRNetwork2')
        return o 


class Alignment(nn.Module):
    def __init__(self, mem_dim, conv_dim, mlp_dim):
        super(Alignment, self).__init__()
        self.word_trans = nn.Linear(4*conv_dim,mem_dim, bias=False)
        self.mlp = nn.Linear(2*mem_dim, mlp_dim)
        self.output = nn.Linear(mlp_dim, 3) # 0: match, 1: umatch, 2: unknown


        self.alignment = nn.Linear(mem_dim + conv_dim*4, 1)

        self.mlp = nn.Linear(mem_dim + conv_dim*4, mlp_dim)

    def forward(self, ehr_vector, criteria, ec_mask):
        # ehr_vector: bs2, mem_dim    << EHR 
        ## ehr_vector 128,320;  
        # criteria: bs1, num_ec, 4*conv_dim,    << from criteria
        ## criteria 10,80,512;  
        # ec_mask: bs1, num_ec 
        ## attention over num_ec

        ## output bs1,bs2,num_ec,3 

        # print(ec_mask.shape, criteria.shape, ehr_vector.shape, 'ec_mask.shape')
        bs1, num_ec, query_dim = criteria.shape 
        bs2, mem_dim = ehr_vector.shape 
        ec_mask_extend = ec_mask.unsqueeze(1).unsqueeze(-1)
        ec_mask_extend = ec_mask_extend.repeat(1,bs2,1,1)
        criteria = criteria.unsqueeze(1)
        criteria = criteria.repeat(1,bs2,1,1) ## bs1,bs2,num_ec,xxx
        ehr_vector = ehr_vector.unsqueeze(0)
        ehr_vector = ehr_vector.unsqueeze(2) ### 1,bs2,1,mem_dim 
        ehr_vector = ehr_vector.repeat(bs1,1,num_ec,1)
        attention = torch.cat([criteria, ehr_vector], dim = -1)
        attention = attention.view(-1, query_dim + mem_dim)
        attention = self.alignment(attention) 
        attention = attention.view(bs1,bs2,num_ec) 
        attention = torch.softmax(attention, dim=2)
        attention = attention.unsqueeze(-1)

        h = torch.cat([criteria, ehr_vector], dim = -1) ### bs1,bs2,num_ec,xxx
        h = h * attention 
        h = h.view(-1, query_dim + mem_dim)
        h = self.mlp(h)
        h = h.view(bs1,bs2,num_ec,-1)

        return h, None, None 





import math
import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter
from torch.nn.modules.module import Module
torch.manual_seed(3) 
np.random.seed(1)

class GraphConvolution(Module):
    """
    Simple GCN layer, similar to https://arxiv.org/abs/1609.02907
    """

    def __init__(self, in_features, out_features, bias=True, init='xavier'):
        super(GraphConvolution, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = Parameter(torch.FloatTensor(in_features, out_features))
        if bias:
            self.bias = Parameter(torch.FloatTensor(out_features))
        else:
            self.register_parameter('bias', None)
        if init == 'uniform':
            print("| Uniform Initialization")
            self.reset_parameters_uniform()
        elif init == 'xavier':
            print("| Xavier Initialization")
            self.reset_parameters_xavier()
        elif init == 'kaiming':
            print("| Kaiming Initialization")
            self.reset_parameters_kaiming()
        else:
            raise NotImplementedError

    def reset_parameters_uniform(self):
        stdv = 1. / math.sqrt(self.weight.size(1))
        self.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)

    def reset_parameters_xavier(self):
        nn.init.xavier_normal_(self.weight.data, gain=0.02) # Implement Xavier Uniform
        if self.bias is not None:
            nn.init.constant_(self.bias.data, 0.0)

    def reset_parameters_kaiming(self):
        nn.init.kaiming_normal_(self.weight.data, a=0, mode='fan_in')
        if self.bias is not None:
            nn.init.constant_(self.bias.data, 0.0)

    def forward(self, input, adj):
        support = torch.mm(input, self.weight)
        output = torch.spmm(adj, support)
        if self.bias is not None:
            return output + self.bias
        else:
            return output

    def __repr__(self):
        return self.__class__.__name__ + ' (' \
               + str(self.in_features) + ' -> ' \
               + str(self.out_features) + ')'


class GraphAttention(nn.Module):
    """
    Simple GAT layer, similar to https://arxiv.org/abs/1710.10903
    """

    def __init__(self, in_features, out_features, dropout, alpha, concat=True):
        super(GraphAttention, self).__init__()
        self.dropout = dropout
        self.in_features = in_features
        self.out_features = out_features
        self.alpha = alpha
        self.concat = concat

        self.W = nn.Parameter(nn.init.xavier_normal_(torch.Tensor(in_features, out_features).type(torch.cuda.FloatTensor if torch.cuda.is_available() else torch.FloatTensor), gain=np.sqrt(2.0)), requires_grad=True)
        self.a1 = nn.Parameter(nn.init.xavier_normal_(torch.Tensor(out_features, 1).type(torch.cuda.FloatTensor if torch.cuda.is_available() else torch.FloatTensor), gain=np.sqrt(2.0)), requires_grad=True)
        self.a2 = nn.Parameter(nn.init.xavier_normal_(torch.Tensor(out_features, 1).type(torch.cuda.FloatTensor if torch.cuda.is_available() else torch.FloatTensor), gain=np.sqrt(2.0)), requires_grad=True)

        self.leakyrelu = nn.LeakyReLU(self.alpha)

    def forward(self, input, adj):
        h = torch.mm(input, self.W)
        N = h.size()[0]

        f_1 = torch.matmul(h, self.a1)
        f_2 = torch.matmul(h, self.a2)
        e = self.leakyrelu(f_1 + f_2.transpose(0,1))

        zero_vec = -9e15*torch.ones_like(e)
        attention = torch.where(adj > 0, e, zero_vec)
        attention = F.softmax(attention, dim=1)
        attention = F.dropout(attention, self.dropout, training=self.training)
        h_prime = torch.matmul(attention, h)

        if self.concat:
            return F.elu(h_prime)
        else:
            return h_prime

    def __repr__(self):
        return self.__class__.__name__ + ' (' + str(self.in_features) + ' -> ' + str(self.out_features) + ')'



